/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file dist_list.h
 *
 * @brief The header for distributed lists.
 */
#ifndef __DIST_LIST_H__
#define __DIST_LIST_H__

namespace sketo {
  /**
   * @brief A class for distributed list (one-dimensional array).
   *
   */
  template <typename A>
  class dist_list {
    friend class _impl::ls_adpt;
    friend class _impl::dist_list_buffer<A>;
    
  public:
    typedef A element_type;
    
    typedef dist_list<A> InitialType;                                 
    const InitialType &getInitialList() const { return *this; }       
    template <typename NextCollector>                                 
    struct CollectorType { typedef _impl::IdCollector<NextCollector> type; };
    template <typename NextCollector>                                 
    _impl::IdCollector<NextCollector>                                        
    getCollector(NextCollector c) const { return _impl::IdCollector<NextCollector>(c); }

  private:
    _impl::dist_list_buffer<A> *data;

  public:
    dist_list();

    /** @brief A constructor that allocates distributed space */
    dist_list(int size);

    /** @brief A constructor that allocates distributed space and copy from (sequential) data */
    dist_list(int size, const A* seq_data);

    /** @brief A copy constructor
     *
     * Actually, this constructor does "shallow copy", that is, the two lists (this and that)
     * share the actual buffer on memary.  Use clone() in addition if you want "deep copy".
     *
     * @param that the reference to another distributed list
     */
    dist_list(const dist_list<A> &that);

    ~dist_list();

    /** @brief Assignement
     *
     * Actually, this assignment does "shallow copy", that is, the two lists (this and that)
     * share the actual buffer on memary.  Use clone() in addition if you want "deep copy".
     *
     * @param that the reference to another distributed list
     */
    dist_list<A>& operator=(const dist_list<A>& that);

  public:
    /** @brief Return the global size of the list */
    int get_global_size() const { return data->get_global_size(); };
    

    /** @brief Get the value of the specified global index.
     * @param index the global index
     */
    A get(int index) const;

    /** @brief For skeleton implementation.
     * @param i the local index 
     * @param l the buffer index
     */
    A local_get(int i, int l = 0) const { return data->local_get(i, l); }

    /** @brief Quick if-less access in shifts */
    A local_get_safe(int i, int l = 0) const { return local_get(i, l); }
    /** @brief The index from which local_get_safe is available. */
    int get_left_unsafe() const { return 0; }
    /** @brief The reverse index from which local_get_safe is available. */
    int get_right_unsafe() const { return 0; }

    /** @brief Quick if-less access in zips */
    A local_get_safe_quick(int i, int l = 0) const { return local_get(i, l); }
    /** @brief True if zipwobjs did not carry out any materialization (rebalancing). */
    bool did_zip_materialization() const { return false; }
    static const bool contains_zip = false;
    
    void fix_buffer(int l) const { data->fix_buffer(l); }
    A local_get_safe_quick_fixed(int i) const { return data->local_get_safe_quick_fixed(i); }
    
    /** @brief Returns the number of sublists in this list.
     */
    int get_num_bufs() const { return data->bufs.size(); }
    /** @brief Returns the size of l-th sublist in this list.
     * This method is valid then is_dirty() returns true.
     * Check the dirty flag by is_dirty() before using this method,
     * and unset the dirty flag by unset_dirty() if necessary.
     */
    int get_local_size(int l = 0) const { return data->get_local_size(l); } 

    /** @brief If this returns false, then the local/global size information is locally available.
     * If this return true, then use unset_dirty() to unset the dirty flag. 
     */
    bool is_dirty() const { return data->is_dirty(); }
    /** @brief Unsets the dirty flag by doing necessary computation and communication.  
     */
    void unset_dirty() const { data->unset_dirty(); }
    /** @brief Makes this list ready for the method local_get(i, l).
     */
    void make_local_get_ready() const { data->make_local_get_ready(); }

    /** @brief Returns the sizes of all sublists in this list.
     * This method is valid then is_dirty() returns true.
     * Check the dirty flag by is_dirty() before using this method,
     * and unset the dirty flag by unset_dirty() if necessary.
     */
    const std::vector< std::vector<int> > & get_sizes() const { return data->get_sizes(); }
    
    bool has_concrete_list() const { return true; }
    const dist_list<A> &get_concrete_list() const { return *this; }
    
    /** @brief Set a value to the specifed global index.
     * @param index the global index.
     * @param val   value to be set.
     */
    void set(int index, A val);

    /** @brief Create a (deep) copy of the list. */
    dist_list<A> clone() const;

    /** @brief Gather the distributed values to the (sequential) array.
     *
     * This function gathers the distributed values to the array specified by the input.
     * The user should prepare enough space (i.e., get_global_size()).
     *
     * For reasons of performance, user can specify the buffer as NULL on some processes,
     * then this function does not gather the values to the process.
     * For example, this is effective when the gathered values are only used in outputting them
     * from the root process.
     *
     * @param buffer The pointer to the array to which the values are gathered, or NULL.
     */
    void gather(A* buffer) const;

    /** @brief Distributes the given array on the specified process among processors.
     *
     * @param buffer The pointer to the array from which the values are distributed.
     * @param n      The number of items in the buffer
     * @param org    Then rank of the process that has the array to be distributed.
     */
    void distribute(const A* buffer, int n, int org);

    /** @brief Adds the given element to the head of the list.
     *
     * @param a   The element to be added to the head of the list.
     */
    void push_front(const A&a);

    /** @brief Adds the given element to the last of the list.
     *
     * @param a   The element to be added to the last of the list.
     */
    void push_back(const A&a);

    /** @brief Removes and returns the head element of the list.   */
    A pop_front();

    /** @brief Removes and returns the last element of the list.   */
    A pop_back();

    /** @brief Redistribute the elements evenly among processes, to become a uni-list. */
    void rebalance();

    /** @brief Produces a uni-list by redistributing the head n elements evenly among processes.
     * In other words, this is 'take n' in the functional programming. 
     *
     * @param n   The size of the new uni-list, and must be smaller than or equal to the size of the current list. 
     */
    dist_list<A> rebalance(int n) const;

    /**
     * @todo write comments
     */
    void
    print(std::ostream &stream = std::cout) const;


#ifndef __SKETO_NO_FUSION__
    // The following are used when fusion transformation is activated.
  public:
    template <typename B>
    dist_list(const B& that);
    
    template <typename B>
    dist_list<A>& operator=(const B& that);
      
      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const {
        if((void*)data == (void*)target) return 1;
        return 0;
      }

#endif // __SKETO_NO_FUSION__
  };

  namespace _impl {

    /**
     * @brief The main data structure for maintaining distributed list.
     */
    template <typename A>
    class dist_list_buffer {
//       friend class dist_list<A>;
//       friend class _impl::ls_adpt;

    public: // we assume classes in _impl are not usually accessed by end-users
    
      bool dirty;   // if dirty == false then sizes is valid
      std::vector< std::vector<int> > sizes;
      int global_size;

      //A **bufsp;
      
      // void make_bufsp() {
      //   if(bufsp) delete bufsp;
      //   bufsp = new A*[bufs.size()];
      //   for(int l = 0; l < bufs.size(); l++) {
      //     bufsp[l] = &bufs[l][0];
      //   }
      // }
      
      int get_global_size();

      bool is_dirty() const { return dirty; }
      void unset_dirty() { if(dirty) gather_sizes(); }
      void make_local_get_ready() const { }
      
      const std::vector< std::vector<int> > &get_sizes() {
        if(dirty) gather_sizes(); 
        return sizes;
      }

      void gather_sizes();

      void even_resize(int n);

      int ref_count;

      std::vector< BUFFER<A> > bufs;

      // A *buf;

    public:
      static dist_list_buffer<A>* get_buffer(const dist_list<A> &as) { return as.data; }

    public:
      dist_list_buffer(int gsize);
      ~dist_list_buffer();

      A local_get(int i, int l = 0) const { return bufs[l][i]; }

      A*fixed_buf;

      void fix_buffer(int l = 0) { fixed_buf = &bufs[l][0]; }
      
      A local_get_safe_quick_fixed(int i) const { return fixed_buf[i]; }
      
      /* int get_global_size() const { return global_size; } */
      int get_local_size(int l = 0) const { return bufs[l].size(); } 

      /** @brief Increment the reference count */
      void increment_ref_count();

      /** @brief Decrement the reference count and free the allocated memory when the count becomes zero */
      void decrement_ref_count();

    public:      // utility functions
      /**
       * @brief Compute the first local index and the local size of a distributed list.
       * @param global_size  global size of the list
       * @param procs        the number of processes
       * @param rank         the rank of process
       * @param local_begin  first local index will be stored
       * @param local_size   local size will be stored
       */
      static void
      compute_local_array(int global_size, int procs, int rank,
                          int* local_begin, int* local_size);

      /**
       * @brief Compute the rank of process that has the value of specified index.
       * @param global_size  global size of the list
       * @param procs        the number of processes
       * @param index        the index to be checked.
       */
      static int
      compute_rank_from_index(int global_size, int procs, int index);


      /** @brief Translates the given global index into a triple-index (r, l, i)
       * where the rank r has the element pointed by the global index 
       * in the l-th buffer of multi-list representation at the local index i.
       *
       * @param gi   The global index.
       * @param r    The rank who has the element of the global index gi.
       * @param l    The l-th buffer contains the element of the global index gi.
       * @param i    The local index of the element in the buffer.
       */
      void translate_global_index_to_local(int gi, int *r, int *l, int *i) const;

      /** @brief Translates the given global reverse-index (index from the last) into a triple-reverse-index (r, l, i)
       * where the rank r has the element pointed by the global index 
       * in the l-th buffer of multi-list representation at the local reverse-index i.
       *
       * @param gi   The global index.
       * @param r    The rank who has the element of the global index gi.
       * @param l    The l-th buffer contains the element of the global index gi.
       * @param i    The local reverse-index of the element in the buffer 
       *              (i.e., the element is at (s - 1 - i) where s is the size of the buffer).
       */
      void translate_global_index_to_local_rev(int gi, int *r, int *l, int *i) const;

    };

  }

}

#endif // __DIST_LIST_H__
