/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file list_skeletons_noopt.tpp
 * @brief Template implementation of list skeletons without fusion optimization.
 */
#ifndef __LIST_SKELETONS_NOOPT_TPP__
#define __LIST_SKELETONS_NOOPT_TPP__
#include <cassert>

#include "list_skeletons_common.tpp"

namespace sketo {
  //------------------------------------------------------------------------
  // generate
  template <typename F>
  dist_list<typename result_of<F(int)>::type>
  list_skeletons::generate(int global_size, const F& f)
  {
    return _impl::ls_adpt::generate<F, typename result_of<F(int)>::type>(global_size, f);
  }

  template <typename F, typename A>
  dist_list<A>
  _impl::ls_adpt::generate(int global_size,
                           const F& f)
  {
    dist_list<A> ret(global_size);
    
    int lb, ls; _impl::dist_list_buffer<A>::compute_local_array(global_size, sketo::procs, sketo::rank, &lb, &ls);

    int index = lb;
    for (typename BUFFER<A>::iterator it = ret.data->bufs[0].begin(); it != ret.data->bufs[0].end(); it++,index++) {
      *it = f(index);
    }
    
    return ret;
  }

  //------------------------------------------------------------------------
  // map
  template <typename F, typename A>
  dist_list<typename result_of<F(A)>::type>
  list_skeletons::map(F f,
                      const dist_list<A>& as)
  {
    return _impl::ls_adpt::map<F, A, typename result_of<F(A)>::type>(f, as);
  }

  template <typename F, typename A, typename B>
  dist_list<B>
  _impl::ls_adpt::map(const F& f,
                      const dist_list<A>& as)
  {
    dist_list<B> ret(0);
    ret.data->bufs.clear();
    for(int l = 0; l < as.data->bufs.size(); l++) {
      ret.data->bufs.push_back(BUFFER<B>());
      ret.data->bufs.back().resize(as.data->bufs[l].size());

      for (int i = 0; i < as.data->bufs[l].size(); ++i) {
	ret.data->bufs[l][i] = f(as.data->bufs[l][i]);
      }
    }    
    ret.data->sizes = as.data->sizes;
    ret.data->dirty = as.data->dirty;
    ret.data->global_size = as.data->global_size;
    return ret;
  }

  //------------------------------------------------------------------------
  // map_with_index
  template <typename F, typename A>
  dist_list<typename result_of<F(int, A)>::type>
  list_skeletons::map_with_index(const F& f,
                                 const dist_list<A>& as)
  {
    return _impl::ls_adpt::map_with_index<F, A, typename result_of<F(int, A)>::type>(f, as);
  }

  template <typename F, typename A, typename B>
  dist_list<B>
  _impl::ls_adpt::map_with_index(const F& f,
                                 const dist_list<A>& as)
  {
    dist_list<B> ret(0);
    ret.data->bufs.clear();

    std::vector<std::vector<int> > sizes;
    for(int l = 0; l < as.data->bufs.size(); l++) {
      std::vector<int> s;
      sizes.push_back(s);
      sizes.back().resize(sketo::procs);
      int size = as.data->bufs[l].size();
      _impl::transporter::allgather(&size, 1, &(*sizes.back().begin()));
    }

    int index = 0;
    for(int l = 0; l < as.data->bufs.size(); l++) {
      for(int r = 0; r < sketo::procs; r++) {
	if(r == sketo::rank) {
      ret.data->bufs.push_back(BUFFER<B>());
      ret.data->bufs.back().resize(as.data->bufs[l].size());
	  for (int i = 0; i < as.data->bufs[l].size(); ++i) {
	    ret.data->bufs[l][i] = f(index + i, as.data->bufs[l][i]);
	  }
	}
	index += sizes[l][r];
      }
    }    
    ret.data->sizes = as.data->sizes;
    ret.data->dirty = as.data->dirty;
    ret.data->global_size = as.data->global_size;
    return ret;
  }
    
  //------------------------------------------------------------------------
  // zip and zipwith
  template <typename A, typename B>
  dist_list<std::pair<A, B> >
  list_skeletons::zip(const dist_list<A>& as,
                      const dist_list<B>& bs)
  {
    return _impl::ls_adpt::zipwith<sketo::functions::mkpair<A, B>, A, B, std::pair<A, B> >(sketo::functions::mkpair<A, B>(), as, bs);
  }

  template <typename F, typename A, typename B>
  dist_list<typename result_of<F(A,B)>::type>
  list_skeletons::zipwith(const F& f,
                          const dist_list<A>& as,
                          const dist_list<B>& bs)
  {
    return _impl::ls_adpt::zipwith<F, A, B, typename result_of<F(A,B)>::type>(f, as, bs);
  }

  template <typename F, typename A, typename B, typename C>
  dist_list<C>
  _impl::ls_adpt::zipwith(const F& f,
                          const dist_list<A>& as,
                          const dist_list<B>& bs)
  {
    // unilistization
    int na = as.get_global_size();
    int nb = bs.get_global_size();

    int n = std::min(na, nb);

    dist_list<A> ras = as.rebalance(n);
    dist_list<B> rbs = bs.rebalance(n);

    dist_list<C> ret(n);

    for (int i = 0; i < ret.data->bufs[0].size(); ++i) {
      ret.data->bufs[0][i] = f(ras.data->bufs[0][i], rbs.data->bufs[0][i]);
    }
    
    return ret;
  }

  //------------------------------------------------------------------------
  // concatmap
  template <typename F, typename A>
  dist_list<typename result_of<F(A)>::type>
  list_skeletons::concatmap(F f,
                           const dist_list<A>& as)
  {
    return _impl::ls_adpt::concatmap<F, A, typename result_of<F(A)>::type>(f, as);
  }

  template <typename F, typename A, typename B>
  dist_list<B>
  _impl::ls_adpt::concatmap(const F& f,
                            const dist_list<A>& as)
  {
    dist_list<B> ret(0);
    ret.data->bufs.clear();
    for(int l = 0; l < as.data->bufs.size(); l++) {
      ret.data->bufs.push_back(BUFFER<B>());
      for (int i = 0; i < as.data->bufs[l].size(); ++i) {
	f(as.data->bufs[l][i], ret.data->bufs[l]);
      }
    }
    ret.data->dirty = true;
    return ret;
  }

  namespace _impl {
    template <typename P>
    struct FilterCollector {
      const P &p;
      FilterCollector(const P &p) : p(p) { }
      typedef typename P::argument_type argument_type;
      typedef argument_type arg1_type;
      typedef argument_type result_type;
      template <typename COLLECTOR>
      void operator()(const argument_type &a, COLLECTOR &c) const
      {
	if (p(a)) { c.push_back(a); }
      }
    };
  }

  //------------------------------------------------------------------------
  // filter
  template <typename P, typename A>
  dist_list<A>
  list_skeletons::filter(P p,
                         const dist_list<A>& as)
  {
    return _impl::ls_adpt::concatmap<_impl::FilterCollector<P>, A, A>(_impl::FilterCollector<P>(p), as);
  }

  //------------------------------------------------------------------------
  // reduce
  template <typename OPLUS, typename A>
  A
  list_skeletons::reduce(const OPLUS &oplus,
			 const A &identity,
                         const dist_list<A>& as)
  {
    return _impl::ls_adpt::reduce(oplus, identity, as);
  }

  template <typename OPLUS, typename A>
  A
  list_skeletons::reduce(const OPLUS &oplus,
                         const dist_list<A>& as)
  {
    return _impl::ls_adpt::reduce(oplus, identity_element(oplus), as);
  }

  template <typename OPLUS, typename A>
  A
  _impl::ls_adpt::reduce(const OPLUS &oplus,
			 const A &identity,
                         const dist_list<A> &as)
  {
    // local reduce
    std::vector<A> vals;
    for(int l = 0; l < as.data->bufs.size(); l++) {
      A val = identity;
      for(int i = 0; i < as.data->bufs[l].size(); i++) {
	val = oplus(val, as.data->bufs[l][i]);
      }
      vals.push_back(val);
    }
    // global reduce
    _impl::ls_adpt::global_reduce(oplus, identity, vals.size(), &(*vals.begin()));
    return vals[0];
  }

  //------------------------------------------------------------------------
  // scan
  template <typename OPLUS, typename A>
  dist_list<A>
  list_skeletons::scan(const OPLUS &oplus,
		       const A &identity,
                       A e,
                       const dist_list<A>& as,
                       A *last_val)
  {
    return _impl::ls_adpt::scan<OPLUS, A>(oplus, identity, e, as, last_val);
  }

  template <typename OPLUS, typename A>
  dist_list<A>
  list_skeletons::scan(const OPLUS &oplus,
                       A e,
                       const dist_list<A>& as,
                       A *last_val)
  {
    return _impl::ls_adpt::scan<OPLUS, A>(oplus, identity_element(oplus), e, as, last_val);
  }

  template <typename OPLUS, typename A>
  dist_list<A>
  _impl::ls_adpt::scan(const OPLUS& oplus,
		       const A &identity,
                       A e,
                       const dist_list<A> &as,
                       A* last_val)
  {
    // local reduce
    std::vector<A> total_sums;
    {
      for(int l = 0; l < as.data->bufs.size(); l++) {
	A total_sum = identity;
	for(int i = 0; i < as.data->bufs[l].size(); i++) {
	  total_sum = oplus(total_sum, as.data->bufs[l][i]);
	}
	total_sums.push_back(total_sum);
      }
    }

    // global scan
    std::vector<A> first_vals(total_sums.size());
    global_scan(oplus, identity, total_sums.size(), &(*first_vals.begin()), &(*total_sums.begin()));

    for(int l = 0; l < total_sums.size(); l++) {
      if (rank == 0 && l == 0) {
	first_vals[l] = e;
      } else {
	first_vals[l] = oplus(e, first_vals[l]);
      }
    }
    
    // share the last value (= total_sum)
    if (last_val) {
      *last_val = oplus(e, total_sums.back());
    }

    dist_list<A> bs(0);
    // local scan
    {
      bs.data->bufs.clear();
      for(int l = 0; l < total_sums.size(); l++) {
      bs.data->bufs.push_back(BUFFER<A>());
      bs.data->bufs.back().resize(as.data->bufs[l].size());
	if(as.data->bufs[l].size() == 0) continue;
	bs.data->bufs.back()[0] = first_vals[l];
	for (int i = 1; i < as.data->bufs[l].size(); ++i) {
	  bs.data->bufs.back()[i] = oplus(bs.data->bufs.back()[i-1], as.data->bufs[l][i-1]);
	}
      }
    }
    bs.data->sizes = as.data->sizes;
    bs.data->dirty = as.data->dirty;
    bs.data->global_size = as.data->global_size;
    
    return bs;
  }

  //------------------------------------------------------------------------
  // scanr
  template <typename OPLUS, typename A>
  dist_list<A>
  list_skeletons::scanr(const OPLUS &oplus,
			const A &identity,
                        A e,
                        const dist_list<A>& as,
                        A *first_val)
  {
    return _impl::ls_adpt::scanr<OPLUS, A>(oplus, identity, e, as, first_val);
  }

  template <typename OPLUS, typename A>
  dist_list<A>
  list_skeletons::scanr(const OPLUS &oplus,
                        A e,
                        const dist_list<A>& as,
                        A *first_val)
  {
    return _impl::ls_adpt::scanr<OPLUS, A>(oplus, identity_element(oplus), e, as, first_val);
  }

  template <typename OPLUS, typename A>
  dist_list<A>
  _impl::ls_adpt::scanr(const OPLUS& oplus,
			const A &identity,
                        A e,
                        const dist_list<A> &as,
                        A* first_val)
  {
    // local reduce
    std::vector<A> total_sums;
    {
      for(int l = 0; l < as.data->bufs.size(); l++) {
	A total_sum = identity;
	for(int i = 0; i < as.data->bufs[l].size(); i++) {
	  total_sum = oplus(total_sum, as.data->bufs[l][i]);
	}
	total_sums.push_back(total_sum);
      }
    }

    // global scan
    std::vector<A> last_vals(total_sums.size());
    global_scanr(oplus, identity, total_sums.size(), &(*last_vals.begin()), &(*total_sums.begin()));

    for(int l = 0; l < total_sums.size(); l++) {
      if (rank == sketo::procs - 1 && l == total_sums.size() - 1) {
	last_vals[l] = e;
      } else {
	last_vals[l] = oplus(last_vals[l], e);
      }
    }
    
    // share the first value (= total_sum)
    if (first_val) {
      *first_val = oplus(total_sums[0], e);
    }

    dist_list<A> bs(0);
    // local scan
    {
      bs.data->bufs.clear();
      for(int l = 0; l < total_sums.size(); l++) {
      bs.data->bufs.push_back(BUFFER<A>());
      bs.data->bufs.back().resize(as.data->bufs[l].size());
	if(as.data->bufs[l].size() == 0) continue;
	bs.data->bufs.back().back() = last_vals[l];
	for (int i = as.data->bufs[l].size() - 2; i >= 0 ; --i) {
	  bs.data->bufs.back()[i] = oplus(as.data->bufs[l][i+1], bs.data->bufs.back()[i+1]);
	}
      }
    }

    bs.data->sizes = as.data->sizes;
    bs.data->dirty = as.data->dirty;
    bs.data->global_size = as.data->global_size;
    return bs;
  }
  
  //------------------------------------------------------------------------
  // postscan
  template <typename OPLUS, typename A>
  dist_list<A>
  list_skeletons::postscan(const OPLUS &oplus,
			   const A &identity,
                           const dist_list<A>& as)
  {
    return _impl::ls_adpt::postscan<OPLUS, A>(oplus, identity, as);
  }

  template <typename OPLUS, typename A>
  dist_list<A>
  list_skeletons::postscan(const OPLUS &oplus,
                           const dist_list<A>& as)
  {
    return _impl::ls_adpt::postscan<OPLUS, A>(oplus, identity_element(oplus), as);
  }
  //------------------------------------------------------------------------
  // postscanr
  template <typename OPLUS, typename A>
  dist_list<A>
  list_skeletons::postscanr(const OPLUS &oplus,
			    const A &identity,
                            const dist_list<A>& as)
  {
    return _impl::ls_adpt::postscanr<OPLUS, A>(oplus, identity, as);
  }

  template <typename OPLUS, typename A>
  dist_list<A>
  list_skeletons::postscanr(const OPLUS &oplus,
                            const dist_list<A>& as)
  {
    return _impl::ls_adpt::postscanr<OPLUS, A>(oplus, identity_element(oplus), as);
  }

  

  //------------------------------------------------------------------------
  // shiftl (shift to left)
  template <typename A>
  dist_list<A>
  list_skeletons::shiftl(const A &rightmost,
                         const dist_list<A> &as,
                         A *leftmost)
  {
    return _impl::ls_adpt::shiftl(rightmost, as, leftmost);
  }

  template <typename A>
  dist_list<A>
  _impl::ls_adpt::shiftl(const A &rightmost,
                         const dist_list<A> &as,
                         A *leftmost)
  {
    dist_list<A> ret = as.clone();
    A a = ret.pop_front();
    if (leftmost) *leftmost = a;
    ret.push_back(rightmost);
    return ret;
  }


  //------------------------------------------------------------------------
  // append
  template <typename A>
  dist_list<A>
  list_skeletons::append(const dist_list<A> &as,
                         const dist_list<A> &bs)
  {
    return _impl::ls_adpt::append(as, bs);
  }

  template <typename A>
  dist_list<A>
  _impl::ls_adpt::append(const dist_list<A> &as,
                         const dist_list<A> &bs)
  {
    dist_list<A> ret(0);
    ret.data->bufs.clear();
    ret.data->bufs.reserve(as.data->bufs.size() + bs.data->bufs.size());
    for(int l = 0; l < as.data->bufs.size(); l++) ret.data->bufs.push_back(as.data->bufs[l]);
    for(int l = 0; l < bs.data->bufs.size(); l++) ret.data->bufs.push_back(bs.data->bufs[l]);

    if(!as.is_dirty() && !bs.is_dirty()) {
      ret.data->dirty = false;
      ret.data->sizes.clear();
      std::vector< std::vector<int> > asizes = as.get_sizes();
      std::vector< std::vector<int> > bsizes = bs.get_sizes();
      for(int i = 0; i < asizes.size(); i++) ret.data->sizes.push_back(asizes[i]);
      for(int i = 0; i < bsizes.size(); i++) ret.data->sizes.push_back(bsizes[i]);
      ret.data->global_size = as.get_global_size() + bs.get_global_size();  
    } else {
      ret.data->dirty = true;
    }
    return ret;
  }



  //------------------------------------------------------------------------
  // shiftr (shift to right)
  template <typename A>
  dist_list<A>
  list_skeletons::shiftr(const A &leftmost,
                         const dist_list<A> &as,
                         A* rightmost)
  {
    return _impl::ls_adpt::shiftr(leftmost, as, rightmost);
  }

  template <typename A>
  dist_list<A>
  _impl::ls_adpt::shiftr(const A &leftmost,
                         const dist_list<A> &as,
                         A *rightmost)
  {
    dist_list<A> ret = as.clone();
    A a = ret.pop_back();
    if (rightmost) *rightmost = a;
    ret.push_front(leftmost);
    return ret;

  }

  //------------------------------------------------------------------------
  // gscanl
  template <typename F, typename G, typename OMINUS, typename OPLUS, typename B, typename A>
  dist_list<B>
  list_skeletons::gscanl(const F &f,
                         const B &c,
                         const G &g,
                         const OMINUS &ominus,
                         const OPLUS &oplus,
			 const typename OPLUS::result_type &identity,
                         const dist_list<A>& as)
  {
    return _impl::ls_adpt::gscanl<F, G, OMINUS, OPLUS, A, B, typename result_of<G(A)>::type>(f, c, g, ominus, oplus, identity, as);
  }

  template <typename F, typename G, typename OMINUS, typename OPLUS, typename B, typename A>
  dist_list<B>
  list_skeletons::gscanl(const F &f,
                         const B &c,
                         const G &g,
                         const OMINUS &ominus,
                         const OPLUS &oplus,
                         const dist_list<A>& as)
  {
    return _impl::ls_adpt::gscanl<F, G, OMINUS, OPLUS, A, B, typename result_of<G(A)>::type>(f, c, g, ominus, oplus, identity_element(oplus), as);
  }


  //------------------------------------------------------------------------
  // gscanr
  template <typename F, typename G, typename OMINUS, typename OPLUS, typename B, typename A>
  dist_list<B>
  list_skeletons::gscanr(const F &f,
                         const B &c,
                         const G &g,
                         const OMINUS &ominus,
                         const OPLUS &oplus,
			 const typename OPLUS::result_type &identity,
                         const dist_list<A>& as)
  {
    return _impl::ls_adpt::gscanr<F, G, OMINUS, OPLUS, A, B, typename result_of<G(A)>::type>(f, c, g, ominus, oplus, identity, as);
  }
  template <typename F, typename G, typename OMINUS, typename OPLUS, typename B, typename A>
  dist_list<B>
  list_skeletons::gscanr(const F &f,
                         const B &c,
                         const G &g,
                         const OMINUS &ominus,
                         const OPLUS &oplus,
                         const dist_list<A>& as)
  {
    return _impl::ls_adpt::gscanr<F, G, OMINUS, OPLUS, A, B, typename result_of<G(A)>::type>(f, c, g, ominus, oplus, identity_element(oplus), as);
  }

}
  
#endif // __LIST_SKELETONS_NOOPT_TPP__
