/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 *
 * @file list_skeletons_predef.h
 * @brief A set of predefinition of class names, etc.
 */
#ifndef __LIST_SKELETONS_PREDEF_H__
#define __LIST_SKELETONS_PREDEF_H__

#include <cassert>

#define BUFFER sketo::_impl::simple_vector

namespace sketo {
  template <typename A> class dist_list;

  namespace _impl {
    template <typename A>
    struct simple_vector {
      A* buf;
      unsigned int tl;
      unsigned int buf_size;

      const static int MINIMUM_BUF_SIZE = 32;

      typedef A* iterator;

      simple_vector() : buf(NULL), tl(0), buf_size(0) {
	reserve( MINIMUM_BUF_SIZE );
      }
      
      simple_vector(int n) : buf(NULL), tl(0), buf_size(0) {
	reserve( std::max(n, MINIMUM_BUF_SIZE) );
	tl = n;
      }

      simple_vector(const simple_vector &that) : buf(NULL), tl(0), buf_size(0){
	reserve( MINIMUM_BUF_SIZE );
	this->operator=(that);
      }

      ~simple_vector() {
	if(buf) delete [] buf;
      }

      const simple_vector &operator=(const simple_vector &that) {
	resize(that.size());
	tl = that.tl;
	for(unsigned int i = 0; i < tl; i++) {
	  buf[i] = that.buf[i];
	}
      }

      A& operator[](int i) const { return buf[i]; }

      A* begin() const { return buf; }
      A* end() const { return buf + tl; }

      A& front() const { return buf[0]; }
      A& back() const { return buf[tl-1]; }

      A& pop_back() { 
	assert (tl > 0) ; 
	tl--; 	
	return buf[tl]; 
      }
      void push_back(const A& a) {
	reserve(tl + 1);

	buf[tl++] = a;
      }

      void insert(iterator it, const A&a) {
	assert( buf <= it && it <= buf + tl );
	unsigned int i = it - buf;
	reserve( tl + 1 );
	for(int j = tl; j > i; j--) buf[j] = buf[j-1];
	buf[i] = a;
	tl++;
      }

      void erase(iterator it) {
	int i = it - buf;
	assert( 0 <= i && i < tl );
	tl--;
	for(int j = i; j < tl; j++) buf[j] = buf[j+1];
      }

      unsigned int size() const { return tl; }

      void resize(unsigned int n) {
	reserve(n);
	tl = n;
      }

      void reserve(unsigned int n) {
	if(n > buf_size) {
	  n = nextsize(n);
	  A *tmp = new A[n];
	  if(buf) {
	    for(unsigned int i = 0; i < tl; i++) {
	      tmp[i] = buf[i];
	    }
	    delete [] buf;
	  }
	  buf = tmp;
	  buf_size = n;
	}
      }

      unsigned int nextsize(unsigned int n) {
	// make the smallest power of 2 not less than n.
	if((n - 1 & n) == 0) return n;  // n is a power of 2
	unsigned int p = 1;
	for(; n > 0; n >>= 1, p <<= 1);
	return p;
      }


    };

    template <typename A> class dist_list_buffer;

    // The terminal collector
    template <typename NextCollector>
    struct IdCollector {
      NextCollector c;
      IdCollector(NextCollector c) : c(c) {}
      template <typename A>
      void push_back(const A& a) {
        //std::cout << "push_back in IdCollector" << std::endl;
        c.push_back(a);
      }
    };
    // Vector-output collector
    template <typename A, typename G>
    struct VectorCollector {
      BUFFER<A> &v;
      const G g;
      VectorCollector(BUFFER<A> &v, const G&g) : v(v),g(g) { }
      void push_back(const typename G::argument_type& a) {
        v.push_back(g(a));
      }
    };
    // Reduce collector
    template<typename A, typename OPLUS, typename G>
    struct ReduceCollector {
      A &res;
      const G g;
      const OPLUS oplus;
      ReduceCollector(A &res, const OPLUS &oplus, const G&g) : oplus(oplus), res(res), g(g) { }
      void push_back(const typename G::argument_type& a) {
        res = oplus(res, g(a));
      }
    };

    template <typename F, typename AS, typename G>
    struct ls_concatmapobj;

    template <typename OP, typename AS, typename A, typename G>
    struct ls_scanobj;

    template <typename OP, typename AS, typename A, typename G>
    struct ls_scanrobj;

    enum {
      TAG_REDUCE,
      TAG_SCAN1,
      TAG_PRINT,
      TAG_SHIFTL,
      TAG_SHIFTR,
      TAG_GATHER,
      TAG_GATHER_SIZE,
    };
    
    class ls_adpt {
    public:
      // driver interfaces
      template <typename COMP, typename AS>
      static void
      dist_list_driver(COMP &comp, const AS &as);

      template <typename COMP, typename AS>
      static void
      dist_list_driver_ow(COMP &comp, const AS &as);

      template <typename COMP, typename AS>
      static void
      dist_list_driver_rev(COMP &comp, const AS &as);

      template <typename COMP, typename AS>
      static void
      dist_list_driver_rev_ow(COMP &comp, const AS &as);

      // static if-branch
      template <bool , typename THEN, typename ELSE>
      struct If {
        typedef ELSE type;
      };
      template <typename THEN, typename ELSE>
      struct If<true, THEN, ELSE> {
        typedef THEN type;
      };

      // driver implementations
      template <typename COMP, typename AS>
      struct dist_list_driver_safe {
        static void drive(COMP &comp, const AS &as);
      };
      template <typename COMP, typename AS>
      struct dist_list_driver_safe_quick_fixed {
        static void drive(COMP &comp, const AS &as);
      };
      template <typename COMP, typename AS>
      struct dist_list_driver_default {
        static void drive(COMP &comp, const AS &as);
      };
      // driver dispatcher
      template <typename COMP, typename AS>
      struct driver_dispacher {
        typedef typename If< AS::contains_zip,
                             dist_list_driver_default<COMP, AS>,
                             dist_list_driver_safe_quick_fixed<COMP, AS> >::type driver;
      };

      template <typename COMP, typename AS>
      struct dist_list_driver_ow_safe {
        static void drive(COMP &comp, const AS &as);
      };
      template <typename COMP, typename AS>
      struct dist_list_driver_ow_safe_quick_fixed {
        static void drive(COMP &comp, const AS &as);
      };
      template <typename COMP, typename AS>
      struct dist_list_driver_ow_default {
        static void drive(COMP &comp, const AS &as);
      };
      // driver dispatcher
      template <typename COMP, typename AS>
      struct driver_ow_dispacher {
        typedef typename If< AS::contains_zip,
                             dist_list_driver_ow_default<COMP, AS>,
                             dist_list_driver_ow_safe_quick_fixed<COMP, AS> >::type driver;
      };

      template <typename COMP, typename AS>
      struct dist_list_driver_rev_safe {
        static void drive(COMP &comp, const AS &as);
      };
      template <typename COMP, typename AS>
      struct dist_list_driver_rev_safe_quick_fixed {
        static void drive(COMP &comp, const AS &as);
      };
      template <typename COMP, typename AS>
      struct dist_list_driver_rev_default {
        static void drive(COMP &comp, const AS &as);
      };
      // driver dispatcher
      template <typename COMP, typename AS>
      struct driver_rev_dispacher {
        typedef typename If< AS::contains_zip,
                             dist_list_driver_rev_default<COMP, AS>,
                             dist_list_driver_rev_safe_quick_fixed<COMP, AS> >::type driver;
      };

      template <typename COMP, typename AS>
      struct dist_list_driver_rev_ow_safe {
        static void drive(COMP &comp, const AS &as);
      };
      template <typename COMP, typename AS>
      struct dist_list_driver_rev_ow_safe_quick_fixed {
        static void drive(COMP &comp, const AS &as);
      };
      template <typename COMP, typename AS>
      struct dist_list_driver_rev_ow_default {
        static void drive(COMP &comp, const AS &as);
      };
      // driver dispatcher
      template <typename COMP, typename AS>
      struct driver_rev_ow_dispacher {
        typedef typename If< AS::contains_zip,
                             dist_list_driver_rev_ow_default<COMP, AS>,
                             dist_list_driver_rev_ow_safe_quick_fixed<COMP, AS> >::type driver;
      };


      template <typename A, typename OP>
      static void
      global_reduce(const OP &oplus,
                    const A &identity,
                    int s,
                    A* pval);

      template <typename A, typename OP>
      static void
      global_scan(const OP &oplus,
                  const A &identity,
                  int s,
                  A* pfirst,
                  A* ptotal);

      template <typename A, typename OP>
      static void
      global_scanr(const OP &oplus,
                   const A &identity,
                   int s,
                   A* plast,
                   A* ptotal);

#ifdef __SKETO_NO_FUSION__
      template <typename F, typename A>
      static dist_list<A>
      generate(int global_size, const F& f);

      template <typename F, typename A, typename B>
      static dist_list<B>
      map(const F& f,
          const dist_list<A>& as);

      template <typename F, typename A, typename B>
      static dist_list<B>
      map_with_index(const F& f,
                     const dist_list<A>& as);
      
      template <typename F, typename A, typename B, typename C>
      static dist_list<C>
      zipwith(const F& f,
              const dist_list<A>& as,
              const dist_list<B>& bs);
      
      template <typename F, typename A, typename B>
      static dist_list<B>
      concatmap(const F& f,
                const dist_list<A>& as);

      template <typename OPLUS, typename A>
      static A
      reduce(const OPLUS &oplus,
             const A &identity,
             const dist_list<A> &as);

      template <typename OPLUS, typename A>
      static dist_list<A>
      scan(const OPLUS& oplus,
           const A &identity,
           A e,
           const dist_list<A> &as,
           A* last_val);

      template <typename OPLUS, typename A>
      static dist_list<A>
      scanr(const OPLUS& oplus,
            const A &identity,
            A e,
            const dist_list<A> &as,
            A* first_val);

      template <typename A>
      static dist_list<A>
      shiftl(const A& rightmost,
             const dist_list<A> &as,
             A* leftmost);

      template <typename A>
      static dist_list<A>
      shiftr(const A& leftmost,
             const dist_list<A> &as,
             A* rightmost);

      template <typename A>
      static dist_list<A>
      append(const dist_list<A> &as,
             const dist_list<A> &bs);

#else
      template <typename OPLUS, typename AS, typename A>
      static A
      reduce(const OPLUS &op, const A &identity, const AS &as);
      
      template <typename OP, typename AS, typename A, typename G, bool overwriting>
      static void
      scanobj_ow_impl(_impl::dist_list_buffer<typename G::result_type>* data, const OP &op, const A &identity, const AS &as, const A &e, A* last, const G&g);

      template <typename A>
      static dist_list_buffer<A> *
      get_dist_list_buffer(const dist_list<A> &as);

      template <typename OP, typename AS, typename A, typename G, bool overwriting>
      static void
      scanrobj_ow_impl(_impl::dist_list_buffer<typename G::result_type>* data, const OP &op, const A &identity, const AS &as, const A &e, A* last, const G&g);
      
      template <typename A>
      static void
      shiftl_edge(const std::vector<A>& old_leftmost,
		  std::vector<A>& new_rightmost,
		  const std::vector< std::vector<int> > &sizes);

      template <typename A>
      static void
      shiftr_edge(const std::vector<A>& old_rightmost,
		  std::vector<A>& new_leftmost,
		  const std::vector< std::vector<int> > &sizes);
    
      template <typename F, typename AS, typename B, typename CMOBJ, typename G>
      static dist_list<B> 
      concatmapobj_materialize(const CMOBJ &cmobj);
      
      template <typename OPLUS, typename FF, typename AS, typename G>
      static typename OPLUS::result_type
      reduce_concatmap(const OPLUS &oplus,
                       const typename OPLUS::result_type &identity,
                       const _impl::ls_concatmapobj<FF, AS, G> &cmobj);

      /*** TENTATIVE TO BE REMOVED   **/
      template <typename A>
      static dist_list<A>
      append(const dist_list<A> &as,
             const dist_list<A> &bs);

#endif

      // For the following four skeletons, we have not
      // implemented fused implementation yet.
      template <typename OPLUS, typename A>
      static dist_list<A>
      postscan(const OPLUS &oplus,
               const A &identity,
               const dist_list<A> &as);

      template <typename OPLUS, typename A>
      static dist_list<A>
      postscanr(const OPLUS &oplus,
                const A &identity,
                const dist_list<A> &as);

      template <typename F, typename G, typename OMINUS, typename OPLUS,
                typename A, typename B, typename C>
      static dist_list<B>
      gscanl(const F &f,
             const B &c,
             const G &g,
             const OMINUS &ominus,
             const OPLUS &oplus,
             const C &identity,
             const dist_list<A> &as);
	
      template <typename F, typename G, typename OMINUS, typename OPLUS,
                typename A, typename B, typename C>
      static dist_list<B>
      gscanr(const F &f,
             const B &c,
             const G &g,
             const OMINUS &ominus,
             const OPLUS &oplus,
             const C &identity,
             const dist_list<A>& as);

    };

  }

}




#endif // __LIST_SKELETONS_PREDEF_H__
