/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file list_skeletons_with_fusion.h
 * @brief An implmentation of list skeletons with fusion optimization.
 */

#ifndef __LIST_SKELETONS_WITH_FUSION_H__
#define __LIST_SKELETONS_WITH_FUSION_H__
#include <cassert>
#include "transporter.h"

#define INSERT_CONCATMAP_FUSION_TERMINAL_DEFINITIONS                            \
  typedef ObjectName InitialType;                                               \
  const InitialType &getInitialList() const { return *this; }                   \
  template <typename NextCollector>                                             \
  struct CollectorType { typedef IdCollector<NextCollector> type; };            \
  template <typename NextCollector>                                             \
  IdCollector<NextCollector>                                                    \
  getCollector(NextCollector c) const { return IdCollector<NextCollector>(c); }
      


namespace sketo {
  namespace _impl {
    
    // Classes for expression templates
    template <typename F>
    struct ls_genobj {
      typedef ls_genobj<F> ObjectName;
      INSERT_CONCATMAP_FUSION_TERMINAL_DEFINITIONS;
      F f;
      int global_size;
      int local_begin;
      int local_size;
      std::vector<std::vector<int> > sizes;
      typedef typename result_of<F(int)>::type element_type;
      ls_genobj(const F &f, int global_size) : f(f), global_size(global_size) {
        sketo::_impl::dist_list_buffer<void>::compute_local_array(global_size, sketo::procs, sketo::rank,
                                                                  &local_begin, &local_size);
	sizes.clear();
        sizes.push_back(std::vector<int>(sketo::procs));
        for(int r = 0; r < sketo::procs; r++) {
          int ls, lb;
          sketo::_impl::dist_list_buffer<void>::compute_local_array(global_size, sketo::procs, r,
                                                                    &lb, &ls);
          sizes[0][r] = ls;
        }
      }
    
      bool has_concrete_list() const { return false; }
      const dist_list<element_type> &get_concrete_list() const { assert( false ); }

      int get_global_size() const { return global_size; }
      int get_local_size(int l = 0) const { return local_size; }
      int get_num_bufs() const { return 1; }
      bool is_dirty() const { return false; }
      void unset_dirty() const { }
      void make_local_get_ready() const { }

      const std::vector<std::vector<int> > &get_sizes() const { return sizes; }

      element_type local_get(int i, int l = 0) const { return f(local_begin + i); }
      element_type local_get_safe(int i, int l = 0) const { return local_get(i, l); }
      int get_left_unsafe() const { return 0; }
      int get_right_unsafe() const { return 0; }
      element_type local_get_safe_quick(int i, int l = 0) const { return local_get_safe(i, l); }
      bool did_zip_materialization() const { return false; }
      static const bool contains_zip = false;
      void fix_buffer(int l) const { }
      element_type local_get_safe_quick_fixed(int i) const { return local_get(i, 0); }

      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const { return 0; }

    };

    template <typename F, typename AS>
    struct ls_mapobj {
      typedef ls_mapobj<F,AS> ObjectName;
      INSERT_CONCATMAP_FUSION_TERMINAL_DEFINITIONS;
      F f;
      const AS as;
      typedef typename result_of<F(typename AS::element_type)>::type element_type;
      ls_mapobj(const F &f, const AS &as) : f(f), as(as) { }
      bool has_concrete_list() const { return false; }
      const dist_list<element_type> &get_concrete_list() const { assert( false ); }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size(int l = 0) const { return as.get_local_size(l); }
      int get_num_bufs() const { return as.get_num_bufs(); }
      bool is_dirty() const { return as.is_dirty(); }
      void unset_dirty() const { as.unset_dirty(); }
      void make_local_get_ready() const { as.make_local_get_ready(); }
      const std::vector<std::vector<int> > &get_sizes() const { return as.get_sizes(); }

      element_type local_get(int i, int l = 0) const { return f(as.local_get(i, l)); }
      element_type local_get_safe(int i, int l = 0) const { return f(as.local_get_safe(i, l)); }
      int get_left_unsafe() const { return as.get_left_unsafe(); }
      int get_right_unsafe() const { return as.get_right_unsafe(); }
      element_type local_get_safe_quick(int i, int l = 0) const { return local_get_safe(i, l); }
      bool did_zip_materialization() const { return as.did_zip_materialization(); }
      static const bool contains_zip = AS::contains_zip;
      void fix_buffer(int l) const { as.fix_buffer(l); }
      element_type local_get_safe_quick_fixed(int i) const { return f(as.local_get_safe_quick_fixed(i)); }

      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const { return as.count_ref(target); }
    };

    template <typename F, typename AS>
    struct ls_mapobj_multi {
      typedef ls_mapobj_multi<F,AS> ObjectName;
      INSERT_CONCATMAP_FUSION_TERMINAL_DEFINITIONS;
      const std::vector<F> fs;
      const AS as;
      typedef typename result_of<F(typename AS::element_type)>::type element_type;
      ls_mapobj_multi(const std::vector<F> &fs, const AS &as) : fs(fs), as(as) { }
      bool has_concrete_list() const { return false; }
      const dist_list<element_type> &get_concrete_list() const { assert( false ); }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size(int l = 0) const { return as.get_local_size(l); }
      int get_num_bufs() const { return as.get_num_bufs(); }
      void is_dirty() const { as.is_dirty(); }
      void unset_dirty() const { as.unset_dirty(); }
      bool make_local_get_ready() const { return as.make_local_get_ready(); }
      const std::vector<std::vector<int> > &get_sizes() const { return as.get_sizes(); }
      element_type local_get(int i, int l = 0) const { return fs[l](as.local_get(i, l)); }
      element_type local_get_safe(int i, int l = 0) const { return fs[l](as.local_get_safe(i, l)); }
      int get_left_unsafe() const { return as.get_left_unsafe(); }
      int get_right_unsafe() const { return as.get_right_unsafe(); }
      element_type local_get_safe_quick(int i, int l = 0) const { return local_get_safe(i, l); }
      bool did_zip_materialization() const { return as.did_zip_materialization(); }
      static const bool contains_zip = AS::contains_zip;

      mutable int _l;
      void fix_buffer(int l) const { as.fix_buffer(l); _l = l; }
      
      element_type local_get_safe_quick_fixed(int i) const { return fs[_l](as.local_get_safe_quick_fixed(i)); }

      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const { return as.count_ref(target); }
    };

    template <typename F, typename AS, typename BS>
    struct ls_zipwobj {
      typedef ls_zipwobj<F,AS,BS> ObjectName;
      INSERT_CONCATMAP_FUSION_TERMINAL_DEFINITIONS;
      F f;
      const AS as;
      const BS bs;

      mutable int global_size;
      mutable dist_list<typename AS::element_type> buf_a; // for rebalancing
      mutable dist_list<typename BS::element_type> buf_b;
      mutable bool rebalanced;
      mutable bool dirty;
      mutable std::vector<std::vector< int > > sizes;
      typedef typename result_of<F(typename AS::element_type, typename BS::element_type)>::type element_type;
      ls_zipwobj(const F &f, const AS &as, const BS &bs) : f(f), as(as), bs(bs) {
        dirty = true;
        rebalanced = false;
        if(!as.is_dirty() && !bs.is_dirty()) {  // sizes are both known 
          int n = 0;
          if(check_sizes(as.get_sizes(), bs.get_sizes(), &n)) {
            dirty = false;
            sizes = as.get_sizes();
            global_size = n;
          }
        } 
      }
      bool has_concrete_list() const { return false; }
      const dist_list<element_type> &get_concrete_list() const { assert( false ); }
      int get_global_size() const { assert (!dirty); return global_size; }
      int get_local_size(int l = 0) const { return get_sizes()[l][sketo::rank]; }
      int get_num_bufs() const { return get_sizes().size(); }
      bool is_dirty() const { return dirty; }
      const std::vector<std::vector<int> > &get_sizes() const { 
        if( dirty ) unset_dirty();
        assert ( !dirty );
        return sizes;
      }
      bool check_sizes(const std::vector< std::vector<int> > &asizes,
                       const std::vector< std::vector<int> > &bsizes,
                       int *n) const {
        bool ok = true;
        *n = 0;
        for(int l = 0; l < asizes.size(); l++) {
          for(int r = 0; r < sketo::procs; r++) {
            ok = ok && ( asizes[l][r] == bsizes[l][r] );
            *n += asizes[l][r];
          }
        }
        return ok;
      }
      void unset_dirty() const {
        if( !dirty ) return;
        if(as.is_dirty()) as.unset_dirty();
        if(bs.is_dirty()) bs.unset_dirty();

        int n = 0;
        if(check_sizes(as.get_sizes(), bs.get_sizes(), &n)) {
          dirty = false;
          rebalanced = false; // no rebalancing is needed at this point
          sizes = as.get_sizes();
          global_size = n;
        } else {
          int na = as.get_global_size();
          int nb = bs.get_global_size();
          n = std::min(na, nb);
          if(as.has_concrete_list()) {
            buf_a = as.get_concrete_list().rebalance(n);
          } else {
            buf_a = as;   // materialize to make it ready for rebalancing
            buf_a = buf_a.rebalance(n);
          }
          if(bs.has_concrete_list()) {
            buf_b = bs.get_concrete_list().rebalance(n);
          } else {
            buf_b = bs;   // materialize to make it ready for rebalancing
            buf_b = buf_b.rebalance(n);
          }
          sizes = buf_a.get_sizes();
          dirty = false;
          rebalanced = true;
          global_size = n;
        }
      }
      void make_local_get_ready() const { 
	if( dirty ) unset_dirty(); 
	if(rebalanced) {
	  buf_a.make_local_get_ready();
	  buf_b.make_local_get_ready();
	} else {
	  as.make_local_get_ready();
	  bs.make_local_get_ready();
	}
      }
      element_type local_get(int i, int l = 0) const { 
        if(rebalanced) return f(buf_a.local_get(i, l), buf_b.local_get(i, l));
        return f(as.local_get(i, l), bs.local_get(i, l)); 
      }
      element_type local_get_safe(int i, int l = 0) const { 
        if(rebalanced) return f(buf_a.local_get_safe(i, l), buf_b.local_get_safe(i, l));
        return f(as.local_get_safe(i, l), bs.local_get_safe(i, l)); 
      }
      int get_left_unsafe() const { 
	if(rebalanced) return std::max(buf_a.get_left_unsafe(), buf_b.get_left_unsafe());
	return std::max(as.get_left_unsafe(), bs.get_left_unsafe());
      }
      int get_right_unsafe() const { 
	if(rebalanced) return std::max(buf_a.get_right_unsafe(), buf_b.get_right_unsafe());
	return std::max(as.get_right_unsafe(), bs.get_right_unsafe());
      }
      element_type local_get_safe_quick(int i, int l = 0) const {
        return f(as.local_get_safe_quick(i, l), bs.local_get_safe_quick(i, l)); 
      }
      bool did_zip_materialization() const { return rebalanced || as.did_zip_materialization() || bs.did_zip_materialization(); }
      static const bool contains_zip = true;

      void fix_buffer(int l) const { as.fix_buffer(l); bs.fix_buffer(l); }
      element_type local_get_safe_quick_fixed(int i) const { 
	return f(as.local_get_safe_quick_fixed(i), bs.local_get_safe_quick_fixed(i));
      }

      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const { return as.count_ref(target) + bs.count_ref(target); }
    };

    template <typename AS, typename A>
    struct ls_shiftlobj {
      typedef ls_shiftlobj<AS,A> ObjectName;
      INSERT_CONCATMAP_FUSION_TERMINAL_DEFINITIONS;
      const AS as;
      const A fromR;
      mutable std::vector<A> edges;
      mutable bool edge_ready;
    public:
      typedef A element_type;
    ls_shiftlobj(const AS &as, const A &fromR) : as(as), fromR(fromR), edge_ready(false) {  }

      bool has_concrete_list() const { return false; }
      const dist_list<element_type> &get_concrete_list() const { assert( false ); }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size(int l = 0) const { return as.get_local_size(l); }
      int get_num_bufs() const { return as.get_num_bufs(); }
      bool is_dirty() const { return as.is_dirty(); }
      void unset_dirty() const { as.unset_dirty(); make_edges(); }
      void make_local_get_ready() const { unset_dirty(); }
      const std::vector<std::vector<int> > &get_sizes() const { return as.get_sizes(); }

      void make_edges() const {
	if(edge_ready) return; // already made.
	edge_ready = true;
	std::vector<A> vals(get_num_bufs());
	for(int l = 0; l < vals.size(); l++) {
	  if(get_local_size(l) > 0) {
	    vals[l] = as.local_get(0, l);
	  }
	}
	edges.clear();
	edges.resize(vals.size(), fromR);
	_impl::ls_adpt::shiftl_edge(vals, edges, as.get_sizes());
	MPI_Barrier( MPI_COMM_WORLD );
      }

      element_type local_get(int i, int l = 0) const {
	if(i + 1 < get_local_size(l)) return as.local_get(i + 1, l);
	else                          return edges[l];
      }
      element_type local_get_safe(int i, int l = 0) const { return as.local_get_safe(i + 1, l); }
      int get_left_unsafe() const { return as.get_left_unsafe(); }
      int get_right_unsafe() const { return as.get_right_unsafe() + 1; }
      element_type local_get_safe_quick(int i, int l = 0) const { return local_get_safe(i, l); }
      bool did_zip_materialization() const { return as.did_zip_materialization(); }
      static const bool contains_zip = AS::contains_zip;
      void fix_buffer(int l) const { as.fix_buffer(l); }
      element_type local_get_safe_quick_fixed(int i) const { return as.local_get_safe_quick_fixed(i+1); }

      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const { return as.count_ref(target); }
    };

    template <typename AS, typename A>
    struct ls_shiftrobj {
      typedef ls_shiftrobj<AS,A> ObjectName;
      INSERT_CONCATMAP_FUSION_TERMINAL_DEFINITIONS;
      const AS as;
      const A fromL;
      mutable std::vector<A> edges;
      mutable bool edge_ready;
    public:
      typedef A element_type;
    ls_shiftrobj(const AS &as, const A &fromL) : as(as), fromL(fromL), edge_ready(false) { }

      bool has_concrete_list() const { return false; }
      const dist_list<element_type> &get_concrete_list() const { assert( false ); }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size(int l = 0) const { return as.get_local_size(l); }
      int get_num_bufs() const { return as.get_num_bufs(); }
      bool is_dirty() const { return as.is_dirty(); }
      void unset_dirty() const { as.unset_dirty(); make_edges(); }
      void make_local_get_ready() const { unset_dirty(); }
      const std::vector<std::vector<int> > &get_sizes() const { return as.get_sizes(); }

      void make_edges() const {
	if(edge_ready) return; // already made.
	edge_ready = true;
	std::vector<A> vals(get_num_bufs());
	for(int l = 0; l < vals.size(); l++) {
	  if(get_local_size(l) > 0) {
	    vals[l] = as.local_get(get_local_size(l) - 1, l);
	  }
	}
	edges.clear();
	edges.resize(vals.size(), fromL);
	_impl::ls_adpt::shiftr_edge(vals, edges, as.get_sizes());
	MPI_Barrier( MPI_COMM_WORLD );
      }

      element_type local_get(int i, int l = 0) const {
	if(i > 0) return as.local_get(i - 1, l);
	else      return edges[l];
      }
      element_type local_get_safe(int i, int l = 0) const { return as.local_get_safe(i - 1, l); }
      int get_left_unsafe() const { return as.get_left_unsafe() + 1; }
      int get_right_unsafe() const { return as.get_right_unsafe(); }
      element_type local_get_safe_quick(int i, int l = 0) const { return local_get_safe(i, l); }
      bool did_zip_materialization() const { return as.did_zip_materialization(); }
      static const bool contains_zip = AS::contains_zip;
      void fix_buffer(int l) const { as.fix_buffer(l); }
      element_type local_get_safe_quick_fixed(int i) const { return as.local_get_safe_quick_fixed(i-1); }
      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const { return as.count_ref(target); }
    };

    template <typename OP, typename AS, typename A, typename G>
    struct ls_scanobj {
      typedef ls_scanobj<OP,AS,A,G> ObjectName;
      INSERT_CONCATMAP_FUSION_TERMINAL_DEFINITIONS;
      const OP op;
      const AS as;
      const A e;
      const A identity;
      const G g;

      A* const last;
      typedef sketo::functions::composition<G, std::binder1st<OP> > FS_Type;
      
      typedef typename G::result_type element_type;

      mutable dist_list<element_type> buf;
      mutable bool materialized;
      
      ls_scanobj(const OP &op, const A &identity, const AS &as, const A &e,  A* last, const G& g) : op(op), identity(identity), as(as), e(e), last(last), g(g), materialized(false) { }
      ~ls_scanobj() { } 
      bool has_concrete_list() const { return materialized; }
      const dist_list<element_type> &get_concrete_list() const { return buf; }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size(int l = 0) const { return as.get_local_size(l); }
      int get_num_bufs() const { return as.get_num_bufs(); }
      bool is_dirty() const { return as.is_dirty(); }
      void unset_dirty() const { return as.unset_dirty(); }
      void make_local_get_ready() const { if (!materialized) materialize(); }
      const std::vector<std::vector<int> > &get_sizes() const { return as.get_sizes(); }
      element_type local_get(int i, int l = 0) const {
        return buf.local_get(i, l);
      }
      element_type local_get_safe(int i, int l = 0) const { return buf.local_get_safe(i, l); }
      int get_left_unsafe() const { return buf.get_left_unsafe(); }
      int get_right_unsafe() const { return buf.get_right_unsafe(); }
      element_type local_get_safe_quick(int i, int l = 0) const { return buf.local_get_safe_quick(i, l); }
      bool did_zip_materialization() const { return false; }
      static const bool contains_zip = false;
      void fix_buffer(int l) const { buf.fix_buffer(l); }
      element_type local_get_safe_quick_fixed(int i) const { return buf.local_get_safe_quick_fixed(i); }
      template <bool overwriting>
      void special_impl(_impl::dist_list_buffer<element_type>* data) const {
        _impl::ls_adpt::scanobj_ow_impl<OP, AS, A, G, overwriting>(data, op, identity, as, e, last, g);
      }

      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const { return as.count_ref(target); }
    private:
      void materialize() const {
        buf = dist_list<element_type>(0);
        special_impl<false>(_impl::ls_adpt::get_dist_list_buffer(buf));
        materialized = true;
      }        
    };

    template <typename OP, typename AS, typename A, typename G>
    struct ls_scanrobj {
      typedef ls_scanrobj<OP,AS,A, G> ObjectName;
      INSERT_CONCATMAP_FUSION_TERMINAL_DEFINITIONS;
      const OP op;
      const AS as;
      const A e;
      const A identity;
      const G g;
      A* first;
      typedef typename G::result_type element_type;
      typedef sketo::functions::composition<G, std::binder2nd<OP> > FS_Type;

      mutable dist_list<element_type> buf;
      mutable bool materialized;
      
      ls_scanrobj(const OP &op, const A &identity, const AS &as, const A &e, A* first, const G&g) : op(op), identity(identity), as(as), e(e), first(first), g(g), materialized(false) { }
      ~ls_scanrobj() { }
      bool has_concrete_list() const { return materialized; }
      const dist_list<element_type> &get_concrete_list() const { return buf; }
      int get_global_size() const { return as.get_global_size(); }
      int get_local_size(int l = 0) const { return as.get_local_size(l); }
      int get_num_bufs() const { return as.get_num_bufs(); }
      bool is_dirty() const { return as.is_dirty(); }
      void unset_dirty() const { return as.unset_dirty(); }
      void make_local_get_ready() const { if (!materialized) materialize(); }
      const std::vector<std::vector<int> > &get_sizes() const { return as.get_sizes(); }
      element_type local_get(int i, int l = 0) const { return buf.local_get(i, l); }
      element_type local_get_safe(int i, int l = 0) const { return buf.local_get_safe(i, l); }
      int get_left_unsafe() const { return buf.get_left_unsafe(); }
      int get_right_unsafe() const { return buf.get_right_unsafe(); }
      element_type local_get_safe_quick(int i, int l = 0) const { return buf.local_get_safe(i, l); }
      bool did_zip_materialization() const { return false; }
      static const bool contains_zip = false;
      void fix_buffer(int l) const { buf.fix_buffer(l); }
      element_type local_get_safe_quick_fixed(int i) const { return buf.local_get_safe_quick_fixed(i); }
      template <bool overwriting>
      void special_impl(_impl::dist_list_buffer<element_type>* data) const {
        _impl::ls_adpt::scanrobj_ow_impl<OP, AS, A, G, overwriting>(data, op, identity, as, e, first, g);
      }
      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const { return as.count_ref(target); }
    private:
      void materialize() const {
        buf = dist_list<element_type>(0);
        special_impl<false>(_impl::ls_adpt::get_dist_list_buffer(buf));
        materialized = true;
      }        
    };
      
    template <typename F, typename AS, typename G>
    struct ls_concatmapobj {
      typedef typename G::result_type element_type;

      typedef typename AS::InitialType InitialType;
      const InitialType &getInitialList() const { return as.getInitialList(); }
      template <typename NextCollector>
      struct Collector {
        NextCollector c;
        const F f;
        Collector(const F&f, NextCollector  c) : f(f), c(c) {}
        void push_back(const typename AS::element_type& a) {
          //std::cout << "push_back in concatMapOnj::Collector" << std::endl;
          f(a, c);
        }
      };

      template <typename NextCollector>
      struct CollectorType {
        typedef typename AS::template CollectorType< Collector<NextCollector> >::type  type;
      };  
      template <typename NextCollector>
      typename CollectorType<NextCollector>::type 
      getCollector(NextCollector c) const {
        return as.getCollector(Collector<NextCollector>(f, c));
      }
      const G g;
      const F f;
      const AS as;
      mutable int global_size;
      mutable std::vector<std::vector<int> > sizes;
      mutable dist_list<element_type> buf;
      mutable bool materialized;
      
      ls_concatmapobj(const F &f, const AS &as, const G &g) : f(f), as(as), g(g) {
        materialized = false;
      }
      bool has_concrete_list() const { return materialized; }
      const dist_list<element_type> &get_concrete_list() const { assert( materialized ); return buf; }
      int get_global_size() const { assert( materialized ); return buf.get_global_size(); }
      int get_local_size(int l = 0) const { assert ( materialized ); return buf.get_local_size(l); }
      int get_num_bufs() const { return as.get_num_bufs(); }
      bool is_dirty() const { return !materialized; }
      void unset_dirty() const { make_local_get_ready(); buf.unset_dirty(); }
      void make_local_get_ready() const {
        if (!materialized) materialize();
      }
      void materialize() const {
        buf = ls_adpt::concatmapobj_materialize<F, AS, element_type, ls_concatmapobj<F, AS, G> , G>(*this);
        materialized = true;
      }
      const std::vector<std::vector<int> > &get_sizes() const {
        return buf.get_sizes();
      }
      element_type local_get(int i, int l = 0) const {
        return buf.local_get(i, l);
      }
      element_type local_get_safe(int i, int l = 0) const { return buf.local_get_safe(i, l); }
      int get_left_unsafe() const { return buf.get_left_unsafe(); }
      int get_right_unsafe() const { return buf.get_right_unsafe(); }
      element_type local_get_safe_quick(int i, int l = 0) const { return local_get_safe(i, l); }
      bool did_zip_materialization() const { return false; }
      static const bool contains_zip = false;
      void fix_buffer(int l) const { buf.fix_buffer(l); }
      element_type local_get_safe_quick_fixed(int i) const { return buf.local_get_safe_quick_fixed(i); }
      template <typename T>
      int count_ref(_impl::dist_list_buffer<T> *target) const { return as.count_ref(target); }
    };

//     template <typename OP, typename AS>
//     class ls_postscanobj {
//       const OP op;
//       const AS as;
//     public:
//       typedef typename AS::element_type element_type;
//       ls_postscanobj(const OP &op, const AS &as) : op(op), as(as) { }
//       int get_global_size() const { return as.get_global_size(); }
//     };

//     template <typename OP, typename AS>
//     class ls_postscanrobj {
//       const OP op;
//       const AS as;
//     public:
//       typedef typename AS::element_type element_type;
//       ls_postscanrobj(const OP &op, const AS &as) : op(op), as(as) { }
//       int get_global_size() const { return as.get_global_size(); }
//     };

  }
  
  namespace list_skeletons {
    /**
     * generate :: int -> (int -> a) -> [a]
     */
    template <typename F>
    _impl::ls_genobj<F>
    generate(int global_size,
             const F& f) {
      return _impl::ls_genobj<F>(f, global_size);
    }

    /**
     * map :: (a -> b) -> [a] -> [b]
     */
    template <typename F, typename AS>
    _impl::ls_mapobj<F, AS>
    map(F f,
        const AS& as) {
      return _impl::ls_mapobj<F, AS>(f, as);
    }

    /**
     * Specialization for map-concatmap fusion
     */
    template <typename F, typename FF, typename AS, typename G>
    _impl::ls_concatmapobj<FF, AS, sketo::functions::composition<F, G> >
    map(F f,
        const _impl::ls_concatmapobj<FF, AS, G> &cmobj) {
      return _impl::ls_concatmapobj<FF, AS, sketo::functions::composition<F, G> >(cmobj.f, cmobj.as, sketo::functions::composition<F, G>(f, cmobj.g));
    }

    /**
     * Specialization for map-scan fusion
     */
    template <typename F, typename OPLUS, typename AS, typename A, typename G>
    _impl::ls_scanobj<OPLUS, AS, A, sketo::functions::composition<F, G> >
    map(F f,
        const _impl::ls_scanobj<OPLUS, AS, A, G> &sobj) {
      return _impl::ls_scanobj<OPLUS, AS, A, sketo::functions::composition<F, G> >(sobj.op, sobj.identity, sobj.as, sobj.e, sobj.last, sketo::functions::composition<F, G>(f, sobj.g));
    }

    /**
     * Specialization for map-scanr fusion
     */
    template <typename F, typename OPLUS, typename AS, typename A, typename G>
    _impl::ls_scanrobj<OPLUS, AS, A, sketo::functions::composition<F, G> >
    map(F f,
        const _impl::ls_scanrobj<OPLUS, AS, A, G> &sobj) {
      return _impl::ls_scanrobj<OPLUS, AS, A, sketo::functions::composition<F, G> >(sobj.op, sobj.identity, sobj.as, sobj.e, sobj.first, sketo::functions::composition<F, G>(f, sobj.g));
    }




    
    /**
     * map_with_index :: (int -> a -> b) -> [a] -> [b]
     */
    template <typename F, typename AS>
    _impl::ls_zipwobj<F, _impl::ls_genobj<sketo::functions::identity<int> >, AS >
    map_with_index(const F& f,
                   const AS& as) {
      return _impl::ls_zipwobj<F, _impl::ls_genobj<sketo::functions::identity<int> >, AS >
        (f, _impl::ls_genobj<sketo::functions::identity<int> >(sketo::functions::identity<int>(), as.get_global_size()), as);
    }

    /**
     * zip :: [a] -> [b] -> [(a, b)]
     */
    template <typename AS, typename BS>
    _impl::ls_zipwobj<sketo::functions::mkpair<typename AS::element_type, typename BS::element_type>, AS, BS>
    zip(const AS& as,
        const BS& bs) {
      return _impl::ls_zipwobj<sketo::functions::mkpair<typename AS::element_type, typename BS::element_type>, AS, BS>
        (sketo::functions::mkpair<typename AS::element_type, typename BS::element_type>(), as, bs);
    }

    /**
     * zipwith :: (a -> b -> c) -> [a] -> [b] -> [c]
     */
    template <typename F, typename AS, typename BS>
    _impl::ls_zipwobj<F, AS, BS>
    zipwith(const F& f,
            const AS& as,
            const BS& bs) {
      return _impl::ls_zipwobj<F, AS, BS>(f, as, bs);
    }

    /**
     * reduce :: (a -> a -> a) -> a -> [a] -> [a]
     */
    template <typename OPLUS, typename AS>
    typename AS::element_type
    reduce(const OPLUS &oplus,
           const typename OPLUS::result_type &identity,
           const AS &as) {
      return _impl::ls_adpt::reduce<OPLUS, AS, typename AS::element_type>(oplus, identity, as);
    }

    /**
     * Specialization for reduce-concatmap fusion
     */
    template <typename OPLUS, typename FF, typename AS, typename G>
    typename OPLUS::result_type
    reduce(const OPLUS &oplus,
           const typename OPLUS::result_type &identity,
           const _impl::ls_concatmapobj<FF, AS, G> &cmobj) {
      return _impl::ls_adpt::reduce_concatmap<OPLUS, FF, AS, G>(oplus, identity, cmobj);
    }

    /**
     * scan :: (a -> a -> a) -> a -> (a, [a]) -> ([a], a)
     */
    template <typename OPLUS, typename AS>
    _impl::ls_scanobj<OPLUS, AS, typename AS::element_type, sketo::functions::identity<typename AS::element_type> >
    scan(const OPLUS &oplus,
         const typename OPLUS::result_type &identity,
         typename OPLUS::first_argument_type e,
         const AS& as,
         typename AS::element_type *last_val = NULL) {
      return _impl::ls_scanobj<OPLUS, AS, typename AS::element_type, sketo::functions::identity<typename AS::element_type> >(oplus, identity, as, e, last_val, sketo::functions::identity<typename AS::element_type>());
    }

    /**
     * scanr :: (a -> a -> a) -> a -> ([a], a) -> (a, [a])
     */
    template <typename OPLUS, typename AS>
    _impl::ls_scanrobj<OPLUS, AS, typename AS::element_type, sketo::functions::identity<typename AS::element_type> >
    scanr(const OPLUS &oplus,
          const typename OPLUS::result_type &identity,
          typename OPLUS::second_argument_type e,
          const AS& as,
          typename AS::element_type *first_val = NULL) {
      return _impl::ls_scanrobj<OPLUS, AS, typename AS::element_type, sketo::functions::identity<typename AS::element_type> >(oplus, identity, as, e, first_val, sketo::functions::identity<typename AS::element_type>());
    }

    template <typename OPLUS, typename AS>
    dist_list<typename AS::element_type>
    postscan(const OPLUS &oplus,
             const typename OPLUS::result_type &identity,
             const AS& as) {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::postscan(oplus, identity, as2);
    }

    template <typename OPLUS, typename AS>
    dist_list<typename AS::element_type>
    postscanr(const OPLUS &oplus,
              const typename OPLUS::result_type &identity,
              const AS& as) {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::postscanr(oplus, identity, as2);
    }
    
    /**
     * shiftl :: ([a], a) -> (a, [a])
     */
    template <typename AS, typename A>
    _impl::ls_shiftlobj<AS, A>
    shiftl(const A &rightmost, 
           const AS &as,
           A *leftmost = NULL) {

      // share leftmost if specified
      if (leftmost) {
        assert(false); // TODO
        //*leftmost = as.get(0);
      }

      return _impl::ls_shiftlobj<AS, A>(as, rightmost);
    }
    
    /**
     * shiftr :: (a, [a]) -> ([a], a)
     */
    template <typename AS, typename A>
    _impl::ls_shiftrobj<AS, A>
    shiftr(const A &leftmost,
           const AS &as,
           A *rightmost = NULL) {

      // share rightmost if specified
      if (rightmost) {
        assert (false) ; // TODO
        //*rightmost = as.get(as.get_global_size()-1);
      }

      return _impl::ls_shiftrobj<AS, A>(as, leftmost);
    }

    /**
     * gscanl :: (b -> a -> b) -> b -> (a -> c) -> (b -> c -> b) -> (c -> c -> c) -> c -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS,
              typename B, typename AS>
    dist_list<B>
    gscanl(const F &f,
           const B &c,
           const G &g,
           const OMINUS &ominus,
           const OPLUS &oplus,
           const typename OPLUS::result_type &identity,
           const AS& as)
    {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::gscanl<F, G, OMINUS, OPLUS, typename AS::element_type, B, typename result_of<G(typename AS::element_type)>::type>(f, c, g, ominus, oplus, identity, as2);
    }

    /**
     * gscanr :: (b -> a -> b) -> b -> (a -> c) -> (b -> c -> b) -> (c -> c -> c) -> c -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS,
              typename B, typename AS>
    dist_list<B>
    gscanr(const F &f,
           const B &c,
           const G &g,
           const OMINUS &ominus,
           const OPLUS &oplus,
           const typename OPLUS::result_type &identity,
           const AS &as)
    {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::gscanr<F, G, OMINUS, OPLUS, typename AS::element_type, B, typename result_of<G(typename AS::element_type)>::type>(f, c, g, ominus, oplus, identity, as2);
    }


    /**
     * concatmap :: (a -> [b]) -> [a] -> [b]
     */
    template <typename F, typename AS>
    _impl::ls_concatmapobj<F, AS, sketo::functions::identity<typename F::result_type> >
    concatmap(F f,
              const AS& as)
    {
      return _impl::ls_concatmapobj<F, AS, sketo::functions::identity<typename F::result_type> >(f, as, sketo::functions::identity<typename F::result_type>());
    }

    template <typename P>
    struct FilterFunction {
      const P p;
      FilterFunction(const P &p) : p(p) { }
      typedef typename P::argument_type argument_type;
      typedef argument_type arg1_type;
      typedef argument_type result_type;
      template <typename COLLECTOR>
      void operator()(const argument_type &a, COLLECTOR &c) const
        {
          if (p(a)) { c.push_back(a); }
        }
    };

    /**
     * filter :: (a -> bool) -> [a] -> [a]
     */
    template <typename P, typename AS>
    _impl::ls_concatmapobj<FilterFunction<P>, AS, sketo::functions::identity<typename AS::element_type> >
    filter(P p,
           const AS& as)
    {
      return concatmap<FilterFunction<P>, AS>(FilterFunction<P>(p), as);
    }



    /**
     * reduce :: (a -> a -> a) -> a -> [a] -> [a]
     */
    template <typename OPLUS, typename AS>
    typename AS::element_type
    reduce(const OPLUS &oplus,
           const AS &as) {
      return _impl::ls_adpt::reduce<OPLUS, AS, typename AS::element_type>(oplus, identity_element(oplus), as);
    }

    /**
     * Specialization for reduce-concatmap fusion
     */
    template <typename OPLUS, typename FF, typename AS, typename G>
    typename OPLUS::result_type
    reduce(const OPLUS &oplus,
           const _impl::ls_concatmapobj<FF, AS, G> &cmobj) {
      return _impl::ls_adpt::reduce_concatmap<OPLUS, FF, AS, G>(oplus, identity_element(oplus), cmobj);
    }

    /**
     * scan :: (a -> a -> a) -> a -> (a, [a]) -> ([a], a)
     */
    template <typename OPLUS, typename AS>
    _impl::ls_scanobj<OPLUS, AS, typename AS::element_type, sketo::functions::identity<typename AS::element_type> >
    scan(const OPLUS &oplus,
         typename OPLUS::first_argument_type e,
         const AS& as,
         typename AS::element_type *last_val = NULL) {
      return _impl::ls_scanobj<OPLUS, AS, typename AS::element_type, sketo::functions::identity<typename AS::element_type> >(oplus, identity_element(oplus), as, e, last_val, sketo::functions::identity<typename AS::element_type>());
    }

    /**
     * scanr :: (a -> a -> a) -> a -> ([a], a) -> (a, [a])
     */
    template <typename OPLUS, typename AS>
    _impl::ls_scanrobj<OPLUS, AS, typename AS::element_type, sketo::functions::identity<typename AS::element_type> >
    scanr(const OPLUS &oplus,
          typename OPLUS::second_argument_type e,
          const AS& as,
          typename AS::element_type *first_val = NULL) {
      return _impl::ls_scanrobj<OPLUS, AS, typename AS::element_type, sketo::functions::identity<typename AS::element_type> >(oplus, identity_element(oplus), as, e, first_val, sketo::functions::identity<typename AS::element_type>());
    }

    template <typename OPLUS, typename AS>
    dist_list<typename AS::element_type>
    postscan(const OPLUS &oplus,
             const AS& as) {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::postscan(oplus, identity_element(oplus), as2);
    }

    template <typename OPLUS, typename AS>
    dist_list<typename AS::element_type>
    postscanr(const OPLUS &oplus,
              const AS& as) {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::postscanr(oplus, identity_element(oplus), as2);
    }



    /***** TENTATIVE TO BE REMOVED ****/
    /**
     * append :: [a] -> [a] -> [a]
     */
    template <typename A>
    dist_list<A>
    append(const dist_list<A> &as,
           const dist_list<A> &bs);


    /**
     * gscanl :: (b -> a -> b) -> b -> (a -> c) -> (b -> c -> b) -> (c -> c -> c) -> c -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS,
              typename B, typename AS>
    dist_list<B>
    gscanl(const F &f,
           const B &c,
           const G &g,
           const OMINUS &ominus,
           const OPLUS &oplus,
           const AS& as)
    {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::gscanl<F, G, OMINUS, OPLUS, typename AS::element_type, B, typename result_of<G(typename AS::element_type)>::type>(f, c, g, ominus, oplus, identity_element(oplus), as2);
    }

    /**
     * gscanr :: (b -> a -> b) -> b -> (a -> c) -> (b -> c -> b) -> (c -> c -> c) -> c -> [a] -> [b]
     */
    template <typename F, typename G, typename OMINUS, typename OPLUS,
              typename B, typename AS>
    dist_list<B>
    gscanr(const F &f,
           const B &c,
           const G &g,
           const OMINUS &ominus,
           const OPLUS &oplus,
           const AS &as)
    {
      dist_list<typename AS::element_type> as2 = as;
      return _impl::ls_adpt::gscanr<F, G, OMINUS, OPLUS, typename AS::element_type, B, typename result_of<G(typename AS::element_type)>::type>(f, c, g, ominus, oplus, identity_element(oplus), as2);
    }

  }
}

#include "list_skeletons_with_fusion.tpp"

#endif // __LIST_SKELETONS_WITH_FUSION_H__


