/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file matrix_skeletons_with_fusion.h
 * @brief An implmentation of matrix skeletons with fusion optimization.
 */

#ifndef __MATRIX_SKELETONS_WITH_FUSION_H__
#define __MATRIX_SKELETONS_WITH_FUSION_H__
#include <cassert>

namespace sketo {
  namespace _impl {
    // Classes for expression templates
    template <typename F>
    struct ms_genobj {
      F f;
      msize global_size;
      mindex local_begin;
      msize local_size;
      typedef typename result_of<F(mindex)>::type element_type;
      ms_genobj(const F &f, msize global_size) : f(f), global_size(global_size) {
        mindexsize mis = sketo::_impl::dist_matrix_buffer<void>::compute_local_array(global_size, sketo::procs, sketo::rank);
        local_begin = mis.index;
        local_size = mis.size;
      }
      msize get_global_size() const { return global_size; }
      msize get_local_size() const { return local_size; }
      element_type local_get(mindex i) const { return f(local_begin + i); }
      template <typename T>
      void check_overwrite(_impl::dist_matrix_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_matrix<element_type>* ow_impl(dist_matrix<element_type>*) const { assert(false); return NULL; }
    };

    template <typename F, typename AS>
    struct ms_mapobj {
      F f;
      const AS as;
      typedef typename result_of<F(typename AS::element_type)>::type element_type;
      ms_mapobj(const F &f, const AS &as) : f(f), as(as) { }
      msize get_global_size() const { return as.get_global_size(); }
      msize get_local_size() const { return as.get_local_size(); }
      element_type local_get(mindex i) const { return f(as.local_get(i)); }
      template <typename T>
      void check_overwrite(_impl::dist_matrix_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_matrix<element_type>* ow_impl(dist_matrix<element_type>*) const { assert(false); return NULL; }
    };

    template <typename F, typename AS, typename BS>
    struct ms_zipwobj {
      F f;
      const AS as;
      const BS bs;
      typedef typename result_of<F(typename AS::element_type, typename BS::element_type)>::type element_type;
      ms_zipwobj(const F &f, const AS &as, const BS &bs) : f(f), as(as), bs(bs) { }
      msize get_global_size() const { return as.get_global_size(); }
      msize get_local_size() const { return as.get_local_size(); }
      element_type local_get(mindex i) const { return f(as.local_get(i), bs.local_get(i)); }
      template <typename T>
      void check_overwrite(_impl::dist_matrix_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_matrix<element_type>* ow_impl(dist_matrix<element_type>*) const { assert(false); return NULL; }
    };
  
    template <typename OP, typename OT, typename AS, typename A>
    struct ms_scanobj {
      const OP op;
      const OT ot;
      const AS as;
      dist_matrix<A> *subarray;
      typedef A element_type;
      ms_scanobj(const OP &op, const OT &ot, const AS &as) : op(op), ot(ot), as(as), subarray(new dist_matrix<A>()) { }
      msize get_global_size() const { return as.get_global_size(); }
      msize get_local_size() const { return as.get_local_size(); }
      element_type local_get(mindex i) const {
        if (_impl::ms_adpt::has_no_data(*subarray)) subeval();
        return subarray->local_get(i);
      }
      ~ms_scanobj() { delete subarray; }
      template <typename T>
      void check_overwrite(_impl::dist_matrix_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_matrix<element_type>* ow_impl(dist_matrix<element_type>* bs) const;
    private:
      void subeval() const {
        dist_matrix<A> array(as.get_global_size());
        *subarray = array;
        _impl::ms_adpt::scanobj_subeval(subarray, op, ot, as);
      }
    };

    template <typename OP, typename OT, typename AS, typename A>
    struct ms_scanrobj {
      const OP op;
      const OT ot;
      const AS as;
      dist_matrix<A> *subarray;
      typedef A element_type;

      ms_scanrobj(const ms_scanrobj<OP, OT, AS, A> &o) : op(op), ot(ot), as(as), subarray(new dist_matrix<A>(*(o.subarray))) {  }
      
      ms_scanrobj(const OP &op, const OT &ot, const AS &as) : op(op), ot(ot), as(as), subarray(new dist_matrix<A>())  { }
      msize get_global_size() const { return as.get_global_size(); }
      msize get_local_size() const { return as.get_local_size(); }
      element_type local_get(mindex i) const {
        if (_impl::ms_adpt::has_no_data(*subarray)) subeval();
        return subarray->local_get(i);
      }
      ~ms_scanrobj() { delete subarray; }
      template <typename T>
      void check_overwrite(_impl::dist_matrix_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_matrix<element_type>* ow_impl(dist_matrix<element_type>*) const;
    private:
      void subeval() const {
        dist_matrix<A> array(as.get_global_size());
        *subarray = array;
        _impl::ms_adpt::scanrobj_subeval(subarray, op, ot, as);
      }
    };
    
    template <int di, int dj, typename F, typename AS, typename A>
    struct ms_shiftobj {
      const F f;
      const AS as;
      typedef A element_type;
      shift_info<di, dj, A> *si;
      ms_shiftobj(const ms_shiftobj<di, dj, F, AS, A> &o) :
        f(o.f), as(o.as), si(new shift_info<di, dj, A>(*(o.si))) { }
      
      ms_shiftobj(const F &f, const AS &as) :
        f(f), as(as), si(new shift_info<di, dj, A>(as.get_global_size())) { }
      msize get_global_size() const { return as.get_global_size(); }
      msize get_local_size() const { return as.get_local_size(); }
      A local_get(const mindex i) const {
        // currently, non-ow implementation is too slow.
        // ow is recommended.
        if(!si->is_ready()) subeval();        
        return si->local_get(i, as);
      }
      ~ms_shiftobj() { delete si; }
      template <typename T>
      void check_overwrite(_impl::dist_matrix_buffer<T>* target,
                           int *refcount, bool *no_deps, bool *child_no_deps, bool *scan_shift) const;
      dist_matrix<A>* ow_impl(dist_matrix<A>*) const;
    private:
      void subeval() const {
        _impl::ms_adpt::shift_common<di, dj, F, AS, A>(NULL, *si, f, as);
        si->gen_index();
      }
    };
  }

  namespace matrix_skeletons {
    /**
     * generate :: msize -> (mindex -> a) -> [a]
     */
    template <typename F>
    _impl::ms_genobj<F>
    generate(msize global_size,
             const F& f) {
      return _impl::ms_genobj<F>(f, global_size);
    }

    /**
     * map :: (a -> b) -> [a] -> [b]
     */
    template <typename F, typename AS>
    _impl::ms_mapobj<F, AS>
    map(F f,
        const AS& as) {
      return _impl::ms_mapobj<F, AS>(f, as);
    }

    /**
     * map_with_index :: (mindex -> a -> b) -> [a] -> [b]
     */
    template <typename F, typename AS>
    _impl::ms_zipwobj<F, _impl::ms_genobj<sketo::functions::identity<mindex> >, AS >
    map_with_index(const F& f,
                   const AS& as) {
      return _impl::ms_zipwobj<F, _impl::ms_genobj<sketo::functions::identity<mindex> >, AS >
        (f, _impl::ms_genobj<sketo::functions::identity<mindex> >(sketo::functions::identity<mindex>(), as.get_global_size()), as);
    }

    /**
     * zip :: [a] -> [b] -> [(a, b)]
     */
    template <typename AS, typename BS>
    _impl::ms_zipwobj<sketo::functions::mkpair<typename AS::element_type, typename BS::element_type>, AS, BS>
    zip(const AS& as,
        const BS& bs) {
      return _impl::ms_zipwobj<sketo::functions::mkpair<typename AS::element_type, typename BS::element_type>, AS, BS>
        (sketo::functions::mkpair<typename AS::element_type, typename BS::element_type>(), as, bs);
    }

    /**
     * zipwith :: (a -> b -> c) -> [a] -> [b] -> [c]
     */
    template <typename F, typename AS, typename BS>
    _impl::ms_zipwobj<F, AS, BS>
    zipwith(const F& f,
            const AS& as,
            const BS& bs) {
      return _impl::ms_zipwobj<F, AS, BS>(f, as, bs);
    }

    /**
     * reduce :: (a -> a -> a) -> (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename OTIMES, typename AS>
    typename AS::element_type
    reduce(const OPLUS &oplus,
           const OTIMES &otimes,
           const AS &as) {
      return _impl::ms_impl_select::ReduceSimple::reduce<OPLUS, OTIMES, AS, typename AS::element_type>(oplus, otimes, as);
    }

    /**
     * reduce :: (a -> a -> a) -> (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename OTIMES, typename AS, typename IMPL>
    typename AS::element_type
    reduce(const OPLUS &oplus,
           const OTIMES &otimes,
           const AS &as,
           const IMPL &impl) {
      return IMPL::template reduce<OPLUS, OTIMES, AS, typename AS::element_type>(oplus, otimes, as);
    }

    /**
     * scan :: (a -> a -> a) -> (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename OTIMES, typename AS>
    _impl::ms_scanobj<OPLUS, OTIMES, AS, typename AS::element_type>
    scan(const OPLUS &oplus,
         const OTIMES &otimes,
         const AS& as) {
      return _impl::ms_scanobj<OPLUS, OTIMES, AS, typename AS::element_type>(oplus, otimes, as);
    }
    
    /**
     * scanr :: (a -> a -> a) -> (a -> a -> a) -> [a] -> [a]
     */
    template <typename OPLUS, typename OTIMES, typename AS>
    _impl::ms_scanrobj<OPLUS, OTIMES, AS, typename AS::element_type>
    scanr(const OPLUS &oplus,
          const OTIMES &otimes,
          const AS& as) {
      return _impl::ms_scanrobj<OPLUS, OTIMES, AS, typename AS::element_type>(oplus, otimes, as);
    }
    
    /**
     * shift :: (Int, Int) -> (((Int, Int), a) -> a) -> [a] -> [a]
     */
    template <int di, int dj, typename F, typename AS>
    _impl::ms_shiftobj<di, dj, F, AS, typename AS::element_type>
    shift(const F &f,
          const AS& as) {
      return _impl::ms_shiftobj<di, dj, F, AS, typename AS::element_type>(f, as);
    }
    
    /**
     * @todo write comments
     */
    template <typename A>
    void
    print(const dist_matrix<A>& as,
          std::ostream &stream = std::cout);

  }
}

#include "matrix_skeletons_with_fusion.tpp"

#endif // __MATRIX_SKELETONS_WITH_FUSION_H__
