/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

#ifndef __SERIALIZATION_H__
#define __SERIALIZATION_H__
namespace sketo {
  namespace serialization {
    /**
     * if you want a struct S to be serialized,
     * please do one of the following.
     *
     * 1. Let the struct S to extend the specific structure:
     *
     *   struct S : sketo::serialization::require_serialization {
     *     // fields
     *   }
     * 
     * 2. Declare the request for serialization in the top namespace.
     *    This is available only for non-template structures (classes).
     *
     *   REQUIRE_SERIALIZATION(S);
     *
     * 3. Declare the specialization of the trait in this namespace.
     *    Suppose S has two template parameters:
     *
     *   namespace sketo {
     *     namespace serialization {
     *       template <typename A, typename B>
     *       struct serialization_trait< S<A, B> > {
     *         static const bool need_serialization = true;
     *       };
     *     }
     *   }
     */
    struct require_serialization {};

    /** meta-if-statement */
    template <bool x, typename T, typename F>
    struct select {
      typedef F value;
    };
    template <typename T, typename F>
    struct select<true, T, F> {
      typedef T value;
    };
    /** a simple predicate to check whether Sub extends Super or not. */
    template <typename Super, typename Sub>
    struct is_superclass_of {
      template<typename T>
      static char (&test(Sub &, T))[1];    
      static char (&test(Super &, int))[2];
      struct C {
        operator Super &() const;
        operator Sub &();
      };
      static const bool value = sizeof(test(C(), 0)) == 1;
    };
    /** a trait to notify the request for serialization. */
    template <typename T>
    struct serialization_trait {
      static const bool need_serialization = is_superclass_of<require_serialization, T>::value;
    };

#define REQUIRE_SERIALIZATION(S)                    \
    namespace sketo {                               \
    namespace serialization {                       \
    tempalte <> struct serialization_trait<S>{      \
      static const bool need_serialization = true;  \
    };                                              \
    }                                               \
    }
    
    enum {
      AR_SIZE_COUNT,
      AR_SERIALIZE,
      AR_DESERIALIZE,
    };
    
    struct size_count {
      static const int ar_type = AR_SIZE_COUNT;
      int size;
      size_count() : size(0) { }
    };
    struct writer {
      static const int ar_type = AR_SERIALIZE;
      unsigned char *buf;
      int size;
      int p;
      writer() : buf(NULL) {};
      void set(int _size) {
        size = _size;
        p = 0;
        buf = new unsigned char[size];
      }
      virtual ~writer() { if(buf!=NULL) delete buf; }
    };
    struct reader {
      static const int ar_type = AR_DESERIALIZE;
      unsigned char *buf;
      int p;
      reader(unsigned char *_buf, int /* size */) : buf(_buf), p(0) { }
    };

    struct access {
      template <typename A, typename Walker>
      static 
      void getinto(const A& a, Walker &w) {
        ((A&)a).serialize(w, 0);
      }
      template <typename A, typename Walker>
      static 
      void getinto(A& a, Walker &w) {
        ((A&)a).serialize(w, 0);
      }
    };
    struct solid_copier {
      template <typename A>
      static 
      void getinto(const A& a, size_count &w) {
        w.size += sizeof(A);
      }
      template <typename A>
      static 
      void getinto(const A& a, writer &w) {
        *((A*)(&w.buf[w.p])) = a;
        w.p += sizeof(A);
      }
      template <typename A>
      static 
      void getinto(A& a, size_count &w) {
        w.size += sizeof(A);
      }
      template <typename A>
      static 
      void getinto(A& a, writer &w) {
        *((A*)(&w.buf[w.p])) = a;
        w.p += sizeof(A);
      }
      template <typename A>
      static 
      void getinto(A& a, reader &w) {
        a = *((A*)(&w.buf[w.p]));
        w.p += sizeof(A);
      }
    };
    template <typename Walker, typename A> 
    void operator&(Walker &w, const A&a) {
      select<sketo::serialization::serialization_trait<A>::need_serialization, access, solid_copier>::value::getinto(a, w);
      //access::getinto(a, w);
    }
    template <typename Walker, typename A> 
    void operator&(Walker &w, A&a) {
      select<sketo::serialization::serialization_trait<A>::need_serialization, access, solid_copier>::value::getinto(a, w);
      //access::getinto(a, w);
    }

#define PRIMITIVE(Walker, Primitive, PrimitiveVar, Operation)   \
    template <>                                                 \
    void operator&<Walker, Primitive>(Walker &w, PrimitiveVar){ \
      Operation;                                                \
    }                                                                 

#define PRIMITIVE_SIZECOUNT(Primitive)                  \
    PRIMITIVE(size_count, Primitive, const Primitive &, \
              w.size += sizeof(Primitive) )             \
    PRIMITIVE(size_count, Primitive, Primitive &,       \
              w.size += sizeof(Primitive) )

    PRIMITIVE_SIZECOUNT(int)
    PRIMITIVE_SIZECOUNT(unsigned int)
    PRIMITIVE_SIZECOUNT(char)
    PRIMITIVE_SIZECOUNT(unsigned char)
    PRIMITIVE_SIZECOUNT(long)
    PRIMITIVE_SIZECOUNT(unsigned long)
    PRIMITIVE_SIZECOUNT(double)
    PRIMITIVE_SIZECOUNT(float)
    PRIMITIVE_SIZECOUNT(bool)

#define PRIMITIVE_WRITE(Primitive)                    \
    PRIMITIVE(writer, Primitive, const Primitive &a,  \
              *((Primitive*)(&w.buf[w.p])) = a;       \
              w.p += sizeof(Primitive) )              \
    PRIMITIVE(writer, Primitive, Primitive &a,        \
              *((Primitive*)(&w.buf[w.p])) = a;       \
              w.p += sizeof(Primitive) )

    PRIMITIVE_WRITE(int)
    PRIMITIVE_WRITE(unsigned int)
    PRIMITIVE_WRITE(char)
    PRIMITIVE_WRITE(unsigned char)
    PRIMITIVE_WRITE(long)
    PRIMITIVE_WRITE(unsigned long)
    PRIMITIVE_WRITE(double)
    PRIMITIVE_WRITE(float)
    PRIMITIVE_WRITE(bool)

#define PRIMITIVE_READ(Primitive)               \
    PRIMITIVE(reader, Primitive, Primitive &a,  \
              a = *((Primitive*)(&w.buf[w.p])); \
              w.p += sizeof(Primitive) )

    PRIMITIVE_READ(int)
    PRIMITIVE_READ(unsigned int)
    PRIMITIVE_READ(char)
    PRIMITIVE_READ(unsigned char)
    PRIMITIVE_READ(long)
    PRIMITIVE_READ(unsigned long)
    PRIMITIVE_READ(double)
    PRIMITIVE_READ(float)
    PRIMITIVE_READ(bool)

    
    struct binary_oarchive {
      writer wr;
      binary_oarchive() {}
      template <typename A>
      void operator<<(const A& a) {
        size_count sc;
        sc & a;
        wr.set(sc.size);
        wr & a;
      }
      unsigned char *buf() { return wr.buf; }
      int size() { return wr.size; }
    };
    struct binary_iarchive {
      reader rr;
      binary_iarchive(unsigned char *buf, int size) : rr(buf, size) { }
      template <typename A>
      void operator>>(A& a) {
        rr & a;
      }
    };
    
    template <typename A>
    struct region_serializer {
      const A*buf;
      int start;
      int skip;
      int len;
      int repeat;
      region_serializer() {}
      region_serializer(const A*_buf, int _start, int _len, int _skip, int _repeat) : buf(_buf), start(_start), skip(_skip), len(_len), repeat(_repeat) { }
    private:
      struct serialize_stub_serialization {
        template<class Archive>
        static
        void serialize(region_serializer &rs, Archive& ar, const unsigned int version)
          {
            A*p = (A*)rs.buf + rs.start;
            for(int i = 0; i < rs.repeat; i++) {
              for(int j = 0; j < rs.len; j++) {
                ar & *p;
                p++;
              }
              p += rs.skip;
            }
          }
      };

      struct serialize_stub_solid {
        static
        void serialize(region_serializer &rs, size_count& ar, const unsigned int version)
          {
            ar.size += sizeof(A) * rs.len * rs.repeat;
          }
        static
        void serialize(region_serializer &rs, writer& ar, const unsigned int version)
          {
            A*p = (A*)rs.buf + rs.start;
            A*q = (A*)(ar.buf + ar.p);
            ar.p += sizeof(A) * rs.len * rs.repeat;
            for(int i = 0; i < rs.repeat; i++) {
              for(int j = 0; j < rs.len; j++) {
                *q++ = *p++;
              }
              p += rs.skip;
            }
          }
        static
        void serialize(region_serializer &rs, reader& ar, const unsigned int version)
          {
            A*p = (A*)rs.buf + rs.start;
            A*q = (A*)(ar.buf + ar.p);
            ar.p += sizeof(A) * rs.len * rs.repeat;
            for(int i = 0; i < rs.repeat; i++) {
              for(int j = 0; j < rs.len; j++) {
                *p++ = *q++;
              }
              p += rs.skip;
            }
          }
      };
      //friend class boost::serialization::access;
      friend class sketo::serialization::access;
      template<class Archive>
      void serialize(Archive& ar, const unsigned int version)
        {
          select<sketo::serialization::serialization_trait<A>::need_serialization, serialize_stub_serialization, serialize_stub_solid>::value::serialize(*this, ar, version);
        }

    };
    template <typename A>
    struct serialization_trait< region_serializer<A> > {
      static const bool need_serialization = true;   // tobe modified.
    };
  }
}
#ifndef __SERIALIZATION_VECTOR_H__
#define __SERIALIZATION_VECTOR_H__
#include <vector>
namespace sketo {
  namespace serialization {

    template <typename A>
    struct serialization_trait< std::vector<A> > {
      static const bool need_serialization = true;
    };

    template <typename A> 
    void operator&(size_count &w, const std::vector<A>&a) {
      const int size = a.size();
      w & size;
      for(int i = 0; i < size; i++) w & *(&a[i]);
    }

    template <typename A> 
    void operator&(writer &w, const std::vector<A>&a) {
      const int size = a.size();
      w & size;
      for(int i = 0; i < size; i++) w & *(&a[i]);
    }
    template <typename A> 
    void operator&(size_count &w, std::vector<A>&a) {
      const int size = a.size();
      w & size;
      for(int i = 0; i < size; i++) w & *(&a[i]);
    }

    template <typename A> 
    void operator&(writer &w, std::vector<A>&a) {
      const int size = a.size();
      w & size;
      for(int i = 0; i < size; i++) w & *(&a[i]);
    }

    template <typename A> 
    void operator&(reader &w, std::vector<A>&a) {
      int size = 0;
      w & size;
      a.resize(size);
      for(int i = 0; i < size; i++) w & *(&a[i]);
    }
  }
}

#endif //__SERIALIZATION_VECTOR_H__

#ifndef __SERIALIZATION_STRING_H__
#define __SERIALIZATION_STRING_H__
#include <string>
#include <cstring>
namespace sketo {
  namespace serialization {

    template <>
    struct serialization_trait< std::string > {
      static const bool need_serialization = true;
    };

    void operator&(size_count &w, const std::string&a) {
      const char *buf = a.c_str();
      int size = strlen(buf);
      w.size += sizeof(int);
      w.size += size;
    }
    void operator&(writer &w, const std::string&a) {
      const char *buf = a.c_str();
      int size = strlen(buf);
      w & size;
      for(int i = 0; i < size; i++) w & buf[i];
    }
    void operator&(size_count &w, std::string&a) {
      const char *buf = a.c_str();
      int size = strlen(buf);
      w.size += sizeof(int);
      w.size += size;
    }
    void operator&(writer &w, std::string&a) {
      const char *buf = a.c_str();
      int size = strlen(buf);
      w & size;
      for(int i = 0; i < size; i++) w & buf[i];
    }
    void operator&(reader &w, std::string&a) {
      int size = 0;
      w & size;
      char *buf = new char[size+1];
      for(int i = 0; i < size; i++) w & buf[i];
      buf[size]=0;
      std::string s(buf);
      a = s;
    }
  }
}

#endif //__SERIALIZATION_STRING_H__

#ifndef __SERIALIZATION_PAIR_H__
#define __SERIALIZATION_PAIR_H__
#include <utility>
namespace sketo {
  namespace serialization {

    template <typename A, typename B>
    struct serialization_trait< std::pair<A, B> > {
      static const bool need_serialization = true;
    };

    template <typename A, typename B> 
    void operator&(size_count &w, const std::pair<A, B>&a) {
      w & a.first;
      w & a.second;
    }

    template <typename A, typename B> 
    void operator&(writer &w, const std::pair<A, B>&a) {
      w & a.first;
      w & a.second;
    }
    template <typename A, typename B> 
    void operator&(size_count &w, std::pair<A, B>&a) {
      w & a.first;
      w & a.second;
    }
    template <typename A, typename B> 
    void operator&(writer &w, std::pair<A, B>&a) {
      w & a.first;
      w & a.second;
    }
    template <typename A, typename B> 
    void operator&(reader &w, std::pair<A, B>&a) {
      w & a.first;
      w & a.second;
    }
  }
}

#endif //__SERIALIZATION_PAIR_H__

#endif //__SERIALIZATION_H__
