#include <iostream>
#include <cmath>
#include <utility>

#include <sketo/list_skeletons.h>

namespace sl = sketo::list_skeletons;

#include "testcase.h"

struct array2D_buffer {
  double data[jmax+2][imax+2];
  int refcount;
  array2D_buffer() : refcount(0) {};
};
  
struct array2D_heap : public sketo::serialization::require_serialization {
   array2D_buffer* buf; // [jmax+2][imax+2];

  array2D_heap() {
    buf = new array2D_buffer(); buf->refcount++;
  }
  array2D_heap(const array2D_heap &x) : buf(x.buf) {
    buf->refcount++;
  }
  array2D_heap& operator=(const array2D_heap &x) {
    if (this->buf == x.buf) return *this;
    this->buf->refcount--; if (this->buf->refcount == 0) delete buf;
    
    this->buf = x.buf; this->buf->refcount++;
    return *this;
  }

  ~array2D_heap() {
    buf->refcount--; if (buf->refcount == 0) delete buf;
  }

private:
  friend class sketo::serialization::access;
  template <class Archive>
  void serialize(Archive& ar, const unsigned int version) {
    assert(this->buf->refcount >= 1);
    if (ar.ar_type == sketo::serialization::AR_DESERIALIZE) {
      if (this->buf) {
	assert(this->buf->refcount >= 1);
	if (this->buf->refcount > 1) {
	  this->buf->refcount--;
	  this->buf = new array2D_buffer();
	  this->buf->refcount = 1;
	}
      } else {
	this->buf = new array2D_buffer();
	this->buf->refcount = 1;
      }
    }

    for (int j = 0; j < jmax + 2; j++) {
      for (int i = 0; i < imax + 2; i++) {
	ar & this->buf->data[j][i];
      }
    }
  }
};

array2D_heap zero_2D;

struct sum_times_2D_t : public sketo::functions::base<double (array2D_heap, array2D_heap)> {
  double operator()(const array2D_heap &x, const array2D_heap &y) const {
    double ret = 0.0;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	ret += x.buf->data[j][i] * y.buf->data[j][i];
      }
    }
    return ret;
  }
} sum_times_2D;

struct minus_times_scalar_2D : public sketo::functions::base<array2D_heap (array2D_heap, array2D_heap)> {
  double scalar;
  minus_times_scalar_2D(double scalar) : scalar(scalar) { }
  array2D_heap operator()(const array2D_heap &x, const array2D_heap &y) const {
    array2D_heap ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	ret.buf->data[j][i] = x.buf->data[j][i] - scalar * y.buf->data[j][i];
      }
    }
    return ret;
  }
};

struct plus_plus_times_scalar_times_scalar : public sketo::functions::base<array2D_heap (std::pair<array2D_heap, std::pair<array2D_heap, array2D_heap> >)> {
  double scalar1, scalar2;
  plus_plus_times_scalar_times_scalar(double scalar1, double scalar2) : scalar1(scalar1), scalar2(scalar2) { }
  array2D_heap operator()(const std::pair<array2D_heap, std::pair<array2D_heap, array2D_heap> > &x) const {
    array2D_heap ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	ret.buf->data[j][i] = x.first.buf->data[j][i] + scalar1 * x.second.first.buf->data[j][i]
                                            + scalar2 * x.second.second.buf->data[j][i];
      }
    }
    return ret;
  }
};


struct times_2D_t : public sketo::functions::base<array2D_heap (array2D_heap, array2D_heap)> {
  array2D_heap operator()(const array2D_heap &x, const array2D_heap &y) const {
    array2D_heap ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	ret.buf->data[j][i] = x.buf->data[j][i] * y.buf->data[j][i];
      }
    }
    return ret;
  }
} times_2D;

struct minus_2D_t : public sketo::functions::base<array2D_heap (array2D_heap, array2D_heap)> {
  array2D_heap operator()(const array2D_heap &x, const array2D_heap &y) const {
    array2D_heap ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	ret.buf->data[j][i] = x.buf->data[j][i] - y.buf->data[j][i];
      }
    }
    return ret;
  }
} minus_2D;

struct sqsum_2D_t : public sketo::functions::base<double (array2D_heap)> {
  double operator()(const array2D_heap &x) const {
    double ret = 0.0;
    for (int j = 1; j <= jmax; j++) {
      for (int i = 1; i <= imax; i++) {
	ret += (x.buf->data[j][i] * x.buf->data[j][i]);
      }
    }
    return ret;
  }
} sqsum_2D;

struct gen_b_t : public sketo::functions::base<array2D_heap (int)> {
  array2D_heap operator()(int k) const {
    array2D_heap ret;
    for (int i = 0; i < imax + 2; i++) {
      for (int j = 0; j < jmax + 2; j++) {
	double v;
	if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1)) {
	  v = 0.0;
	} else {
	  v = 2.0;
	  if ((i > 1) && (i < imax)) v += 2.0; else v += 1.0;
	  if ((j > 1) && (j < jmax)) v += 2.0; else v += 1.0;
	  if ((k > 0) && (k < kmax-1)) v += 2.0; else v += 1.0;
	}
	ret.buf->data[j][i] = v;
      }
    }
    return ret;
  }
} gen_b;

struct get_f_t : public sketo::functions::base<array2D_heap (int)> {
  array2D_heap operator()(int k) const {
    array2D_heap ret;
    for (int i = 0; i < imax + 2; i++) {
      for (int j = 0; j < jmax + 2; j++) {
	if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1)) {
	  ret.buf->data[j][i] = 0.0;
	} else {
	  ret.buf->data[j][i] = 1.0;
	}
      }
    }
    return ret;
  }
} gen_f;
    
void make_vector(double *ap, double *ae, double *aw, double *an, double *as, double *ab, double *at,
		 sketo::dist_list<array2D_heap> *b, sketo::dist_list<array2D_heap> *f)
{
  *b = sl::generate(kmax, gen_b);
  *f = sl::generate(kmax, gen_f);

  *ap = 2.0;
  *ae = *aw = *an = *as = *at = *ab = 1.0;
}

struct local_2D : public sketo::functions::base<array2D_heap (array2D_heap)> {
  double ap, aw, ae, as, an;
  local_2D(double ap, double aw, double ae, double as, double an) :
    ap(ap), aw(aw), ae(ae), as(as), an(an) { }
  array2D_heap operator()(const array2D_heap &x) const {
    array2D_heap ret;
    for (int j = 1; j <= jmax; j++) {
      for (int i = 1; i <= imax; i++) {
	ret.buf->data[j][i] = ap * x.buf->data[j][i] + aw * x.buf->data[j][i-1] + ae * x.buf->data[j][i+1]
	                                   + as * x.buf->data[j-1][i] + an * x.buf->data[j+1][i];
      }
    }
    return ret;
  }
};

struct func1 : public sketo::functions::base<array2D_heap (std::pair<array2D_heap, std::pair<array2D_heap, array2D_heap> >)> {
  double ap, ae, aw, an, as, at, ab;
  func1(double ap, double ae, double aw, double an, double as, double at, double ab) :
    ap(ap), ae(ae), aw(aw), an(an), as(as), at(at), ab(ab) { }
  array2D_heap operator()(const std::pair<array2D_heap, std::pair<array2D_heap, array2D_heap> > &blr) const {
    array2D_heap ret;
    const array2D_heap &b = blr.first;
    const array2D_heap &l = blr.second.first;
    const array2D_heap &r = blr.second.second;
    for (int j = 1; j <= jmax; j++) {
      for (int i =  1; i <= imax; i++) {
	ret.buf->data[j][i] = ap * b.buf->data[j][i] + aw * b.buf->data[j][i-1] + ae * b.buf->data[j][i+1]
	  + as * b.buf->data[j-1][i] + an * b.buf->data[j+1][i]
	  + ab * l.buf->data[j][i] 
	  + at * r.buf->data[j][i];	  
      }
    }
    return ret;
  }
};

void probmv(double ap, double ae, double aw, double an, double as, double at, double ab,
	    const sketo::dist_list<array2D_heap> &b, sketo::dist_list<array2D_heap> *c,
	    const sketo::dist_list<array2D_heap> &f)
{
  *c = sl::zipwith(times_2D, f,
		   sl::map(func1(ap, aw, ae, as, an, ab, at),
// 			   sl::zip(b, sl::zip(b, b))));
			   sl::zip(b, sl::zip(sl::shiftl(zero_2D, b), sl::shiftr(zero_2D, b)))));
}

std::pair<int, double>
bicgstab_band(double ap, double ae, double aw ,double an, double as, double at, double ab,
	      sketo::dist_list<array2D_heap> *x,
	      sketo::dist_list<array2D_heap> *b,
	      const sketo::dist_list<array2D_heap> &f)
{
//   double er0 = EPS;

  sketo::dist_list<array2D_heap> c, r, p, r0, e, y, v;

  probmv(ap, ae, aw, an, as, at, ab, *x, &c, f);

//     for (int k = 0; k < kmax; k++) {
//       array2D_heap tmp = c.local_get(k);
//       for (int j = 1; j <= jmax; j++) {
// 	for (int i = 1; i <= imax; i++) {
// 	  std::cout << " " << tmp.data[j][i];
// 	}
//       }
//       std::cout << std::endl;
//     }
//     std::cout << std::endl;

  r = sl::zipwith(minus_2D, *b, c);
  double c1 = sl::reduce(std::plus<double>(),
			 sl::map(sqsum_2D, r));

//   if (c1 < er0) {
//     return std::pair<int, double>(0, c1);
//   }

  p = r.clone();
  r0 = r.clone();

  std::pair<int, double> ret;
  for (int it = 0; it < ITRMAX; it++) {

    probmv(ap, ae, aw, an, as, at, ab, p, &y, f);

    double c2 = sl::reduce(std::plus<double>(),
			   sl::zipwith(sum_times_2D, r0, y));
    double alp = c1 / c2;

    e = sl::zipwith(minus_times_scalar_2D(alp), r, y);

    probmv(ap, ae, aw, an, as, at, ab, e, &v, f);

    double ev = sl::reduce(std::plus<double>(),
			   sl::zipwith(sum_times_2D, e, v));
    double vv = sl::reduce(std::plus<double>(),
			   sl::zipwith(sum_times_2D, v, v));
    
    double c3 = ev / vv;
    *x = sl::map(plus_plus_times_scalar_times_scalar(alp, c3),
		 sl::zip(*x, sl::zip(p, e)));
    
    r = sl::zipwith(minus_times_scalar_2D(c3), e, v);
      
    double rr = sl::reduce(std::plus<double>(),
			   sl::zipwith(sum_times_2D, r, r));
//     if (std::fabs(rr) < er0) { ret = std::pair<int, double>(it + 1, rr); break; }

    c1 = sl::reduce(std::plus<double>(),
		    sl::zipwith(sum_times_2D, r0, r));
    double bet = c1/(c2 * c3);
    p = sl::map(plus_plus_times_scalar_times_scalar(bet, -bet * c3),
		sl::zip(r, sl::zip(p, y)));
    
    ret = std::pair<int, double>(it + 1, rr);
  }
  return ret;
}

struct gen_x_t : public sketo::functions::base<array2D_heap (int)> {
  array2D_heap operator()(int k) const {
    array2D_heap ret;
    for (int j = 0; j < jmax+2; j++) {
      for (int i = 0; i < imax+2; i++) {
	if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1)) {
	  ret.buf->data[j][i] = 0.0;
	} else {
	  ret.buf->data[j][i] = 1.0 + 0.01;
	}
      }
    }
    return ret;
  }
} gen_x;

int sketo::main(int, char**)
{
  // zero_2D
  for (int j = 0; j < jmax+2; j++) {
    for (int i = 0; i < imax+2; i++) {
      zero_2D.buf->data[j][i] = 0.0;
    }
  }
  
  double ap, ae, aw, an, as, at, ab;
  sketo::dist_list<array2D_heap> b, f;
  make_vector(&ap, &ae, &aw, &an, &as, &at, &ab, &b, &f);

  sketo::dist_list<array2D_heap> x = sl::generate(kmax, gen_x);
  
  double st, et;
  MPI_Barrier(MPI_COMM_WORLD);
  st = MPI_Wtime();
#ifndef PROBMV_ONLY
  std::pair<int, double> result = bicgstab_band(ap, ae, aw, an, as, at, ab,
						&x, &b, f);
  std::cout << "error = " << result.second << std::endl;
#else
  ap = -4.5;
  sketo::dist_list<array2D_heap> y(kmax);
  // avoid too-much optimization
  double tmpval = 0.0;
  for (int i = 0; i < ITRMAX * 2; i++) {
    probmv(ap, ae, aw, an, as, at, ab, x, &y, f);
    tmpval += sketo::_impl::dist_list_buffer<array2D_heap>::get_buffer(y)->buf[0].buf->data[0][0];
  }
  sketo::cout << tmpval << std::endl;
#endif
  MPI_Barrier(MPI_COMM_WORLD);
  et = MPI_Wtime();
  sketo::cout << "time = " << (et - st) << std::endl;
  return 0;
}



