/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file heat_equation.cpp
 * @brief This program compute heat equation.
 * @author: Hideo NISHIMURA (nish@ipl.cs.uec.ac.jp)
 * @author: Kiminori Matsuzaki (kmatsu@ipl.t.u-tokyo.ac.jp; converted to new sketo implementation)
 */

//-----------------------------//
//         How to use          //
//-----------------------------//
//
// This program is solving method of one-dimensional heat equation.
// (  , "partial differential equation" or "boundary value problem" )
//
//  Argument of console command :  %s size number
//     "length" is the length of x-coordinate
//     "time" is the passing time
//
//  After running this program, we  get u(x, t) of each x (interval : 1)
//  u(x, t) is a condition after "time" has passed. 
//
//  
//  We use "0 degrees C with both ends" as boundary value.
//  And initial condtion phi(x) , i.e. u(x, 0), is 30 degrees C ( constant )
//   
//  If you want to change initial condition, rewrite function "double phi( double )"

//-----------------------------//
//           Compile           //
//-----------------------------//
//
//  put C++ parallel library and "heat_equation.cpp" in the same directory
//
//     $ mpiCC heat_equation.cpp
//
//  to execute
//     $ mpirun -np "N" a.out "length" "time"

#include <sketo/list_skeletons.h>
#include <cstdlib>
#include <functional>

//------------------------------------------------//
// global variables ( parameters for calculation) //
//------------------------------------------------//
const double delta_x = 1;           // delta x
const double delta_t = 0.001;       // delta t
const double diffuse = 0.094099517; // heat diffusivity ( of copper )
const double boundary = 0.0;        // boundary condition

//----------------------------------//
//  functions using with Skelton    //
//----------------------------------//

//--- initial condition  i.e. u(x, 0)
// phi(x) returns 30 degrees C for each x
struct phi_t : public sketo::functions::base <double (int)> {
  double operator()(int) const {
    return 30.0;
  }
} phi;

//------------------------//
//     main function      //
//------------------------//
int sketo::main(int argc, char **argv)
{
  if (argc != 3) {
    sketo::cout << "usage: " << argv[0] << " length time" << std::endl;
    return 1;
  }

  const int length = std::atoi(argv[1]); // length of bar
  const int time = std::atoi(argv[2]);   // passed time

//   skeleton::time::initialize( ); // for time keeping

  // Generate a dist_list "bar" which means a condition of the bar.
  // Now, each element of dist_list is an inital value.
  dist_list<double> bar = sketo::list_skeletons::generate(length, phi);
  
  dist_list<double> list1, list2;
  
  int count = 0;
  for (double t = delta_t; count < time / delta_t; t+= delta_t, ++count) {
    // ----- algorithm -----
    // We approximate Heat Equation Problem by Finite Difference Method
    //
    //  bar( x, t+1 ) =
    //    gamma * ( bar( x+1, t ) + bar( x-1, t ) - 2 * bar( x, t ) ) + bar( x, t )
    //
    //        bar( x, t ) : Temperature with location ( X-axis ) "x", time "t"
    //           gamma    : = diffuse * delta_t / delta_x * delta_x
    //          diffuse   : Heat Diffusivity
    //

    // list1 = leftwards shifted array ( for access to immediate right )
    list1 = sketo::list_skeletons::shiftl(boundary, bar);

    // list2 = rightwards shifted array ( for access to immediate left )
    list2 = sketo::list_skeletons::shiftr(boundary, bar);

    // store [ bar( x-1, t ) + bar( x+1, t ) ] into list1
    // store [ bar( x-1, t ) + bar( x+1, t ) - 2*bar( x, t ) ] into list1 
    list1 = sketo::list_skeletons::zipwith(std::plus<double>(),
					   sketo::list_skeletons::zipwith(std::plus<double>(), list1, list2),
					   sketo::list_skeletons::map(std::bind1st(std::multiplies<double>(), -2.0), bar));
    // (bar_mul)

    // multiple dist_list "list1" by gamma
    // store gamma * [ bar( x-1, t ) + bar( x+1, t ) - 2*bar( x, t ) ] + bar( x, t ) into bar.
    bar = sketo::list_skeletons::zipwith(std::plus<double>(),
					 bar,
					 sketo::list_skeletons::map(std::bind1st(std::multiplies<double>(), (diffuse * delta_t)), list1));
  }

//   skeleton::time::set_time( "calculate" ); // for time keeping
#if DEBUG_OUTPUT
  list_skeletons::print( bar );
#endif

  return 0;
}
